# StarPU --- Runtime system for heterogeneous multicore architectures.
#
# Copyright (C) 2020-2023  Universit'e de Bordeaux, CNRS (LaBRI UMR 5800), Inria
#
# StarPU is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or (at
# your option) any later version.
#
# StarPU is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See the GNU Lesser General Public License in COPYING.LGPL for more details.
#

try:
    import numpy as np
except (ModuleNotFoundError, ImportError):
	print("Can't find \"Python3 NumPy\" module (consider running \"pip3 install numpy\" or refer to https://numpy.org/install/)")
	exit(77)

import starpu
from starpu import Handle
import asyncio

try:
        starpu.init()
except Exception as e:
        print(e)
        exit(77)

@starpu.access(a="RW",b="R")
def arr_add(a,b):
	for i in range(np.size(a)):
		a[i] = a[i] + b[i]

a_h = Handle(np.array([1, 2, 3, 4]))
b_h = Handle(np.array([5, 6, 7, 8]))

starpu.task_submit(ret_fut=False)(arr_add, a_h, b_h)

print("Array is", a_h.get())

a_h.unregister()
b_h.unregister()

starpu.shutdown()



# StarPU --- Runtime system for heterogeneous multicore architectures.
#
# Copyright (C) 2020, 2022       Universit'e de Bordeaux, CNRS (LaBRI UMR 5800), Inria
#
# StarPU is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or (at
# your option) any later version.
#
# StarPU is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See the GNU Lesser General Public License in COPYING.LGPL for more details.
#
from starpu import starpupy
import starpu
import inspect

def access(**kwargs):
	def access_decorator(func):
		# write access modes in f.access attribute
		setattr(func,'starpu_access', kwargs)
		return func
	return access_decorator

# set mode as **kwargs of func
def set_access(func, **kwargs):
	# write access modes in f.access attribute
	setattr(func,'starpu_access', kwargs)
	return func





















# StarPU --- Runtime system for heterogeneous multicore architectures.
#
# Copyright (C) 2020-2023  Universite de Bordeaux, CNRS (LaBRI UMR 5800), Inria
#
# StarPU is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or (at
# your option) any later version.
#
# StarPU is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See the GNU Lesser General Public License in COPYING.LGPL for more details.
#
from math import sqrt
import starpu
from starpu import starpupy
import time
import asyncio

def await_fut(fut):
    return fut.result()

try:
        starpu.init()
except Exception as e:
        print(e)
        exit(77)

############################################################################
#function no input no output print hello world
def hello():
	print ("Example 1:")
	print ("Hello, world!")

#############################################################################

#function no input no output
def func1():
	print ("Example 2:")
	print ("This is a function no input no output")

##############################################################################

#using decorator wrap the function no input no output
@starpu.delayed
def func1_deco():
	#time.sleep(1)
	print ("Example 3:")
	print ("This is a function no input no output wrapped by the decorator function")

##############################################################################

#function no input return a value
def func2():
	print ("Example 4:")
	return 12

###############################################################################

#function has 2 int inputs and 1 int output
def multi(a,b):
	print ("Example 5:")
	return a*b
#print(multi(2, 3))

###############################################################################

#function has 4 float inputs and 1 float output
def add(a,b,c,d):
	print ("Example 6:")
	return a+b+c+d
#print(add(1.2, 2.5, 3.6, 4.9))

###############################################################################

#function has 2 int inputs 1 float input and 1 float output 1 int output
def sub(a,b,c):
	print ("Example 7:")
	return a-b-c, a-b
#print(sub(6, 2, 5.9))

###############################################################################

#using decorator wrap the function with input
@starpu.delayed(name="test")
def add_deco(a,b,c):
	#time.sleep(1)
	print ("Example 8:")
	print ("This is a function with input and output wrapped by the decorator function:")
	return a+b+c

###############################################################################

#using decorator wrap the function with input
@starpu.delayed(color=1)
def sub_deco(x,a):
	print ("Example 9:")
	print ("This is a function with input and output wrapped by the decorator function:")
	return x-a

###############################################################################

async def main():
	#submit function "hello"
    fut = starpu.task_submit()(hello)
    await(fut)

    #submit function "func1"
    fut1 = starpu.task_submit()(func1)
    await(fut1)

    #apply starpu.delayed(func1_deco())
    await(func1_deco())

	#submit function "func2"
    fut2 = starpu.task_submit()(func2)
    res2 = await(fut2)
	#print the result of function
    print("This is a function no input and the return value is", res2)

    #submit function "multi"
    fut3 = starpu.task_submit()(multi, 2, 3)
    res3 = await(fut3)
    print("The result of function multi is :", res3)

	#submit function "add"
    fut4 = starpu.task_submit()(add, 1.2, 2.5, 3.6, 4.9)
    res4 = await(fut4)
    print("The result of function add is :", res4)

	#submit function "sub" but only provide function name
    fut5 = starpu.task_submit()(sub, 6, 2, 5.9)
    res5 = await(fut5)
    print("The result of function sub is:", res5)

	#apply starpu.delayed(add_deco)
    fut6 = add_deco(1,2,3)
    #res6 = await(fut6)
    #print("The result of function is", res6)

    #apply starpu.delayed(sub_deco)
    fut7 = sub_deco(fut6, 1)
    res7 = await(fut7)
    print("The first argument of this function is the result of Example 8")
    print("The result of function is", res7)

    fut8 = starpu.task_submit()("sqrt", 4)
    res8 = await(fut8)
    print("The result of function sqrt is:", res8)

try:
        asyncio.run(main())
except starpupy.error as e:
        print("No worker to execute the job")
        starpu.shutdown()
        exit(77)

starpu.shutdown()
#starpu.task_wait_for_all()



# StarPU --- Runtime system for heterogeneous multicore architectures.
#
# Copyright (C) 2020-2023  Universit'e de Bordeaux, CNRS (LaBRI UMR 5800), Inria
#
# StarPU is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or (at
# your option) any later version.
#
# StarPU is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See the GNU Lesser General Public License in COPYING.LGPL for more details.
#

import starpu
from starpu import starpupy
import os
import inspect
import array

# define the different architecture
STARPU_CPU_WORKER = 0
STARPU_CUDA_WORKER = 1
STARPU_OPENCL_WORKER = 2
STARPU_MAX_FPGA_WORKER = 4
STARPU_MPI_MS_WORKER = 5
STARPU_TCPIP_MS_WORKER = 6
STARPU_HIP_WORKER = 7
STARPU_NARCH = 8
STARPU_ANY_WORKER = 255

#class perfmodel
class Perfmodel(object):
	def __init__(self, symbol):
		self.symbol=symbol
		self.pstruct=starpupy.init_perfmodel(self.symbol)

	def get_struct(self):
		return self.pstruct

	def __del__(self):
	#def free_struct(self):
		starpupy.free_perfmodel(self.pstruct)

# generate the dictionary which contains the perfmodel symbol and its struct pointer
dict_perf={}
def dict_perf_generator(perfsymbol):
	if dict_perf.get(perfsymbol)==None:
		p=Perfmodel(perfsymbol)
		dict_perf[perfsymbol]=p
	else:
		p=dict_perf[perfsymbol]
	return p

# add options in function task_submit
def task_submit(**kwargs):
	# set perfmodel
	perf=None
	if kwargs.__contains__("perfmodel") and kwargs['perfmodel']!=None:
		perf=dict_perf_generator(kwargs['perfmodel'])
	kwargs['perfmodel']=perf

	def call_task_submit(f, *args):
		modes={}
		# if there is access mode defined
		if hasattr(f,"starpu_access"):
			# the starpu_access attribute of f is the access mode
			access_mode=f.starpu_access
			# get the name of formal arguments of f
			f_args = inspect.getfullargspec(f).args
			# check the access right of argument is set in mode or not
			for i in range(len(f_args)):
				if f_args[i] in access_mode.keys():
					# set access mode in modes option
					modes[id(args[i])]=access_mode[f_args[i]]
		kwargs['modes']=modes

		res=starpupy._task_submit(f, *args, kwargs)
		return res
	return call_task_submit

# dump performance model and show the plot
def perfmodel_plot(perfmodel, view=True):
	p=dict_perf[perfmodel]
	starpupy.save_history_based_model(p)
	if view == True:
		os.system('starpu_perfmodel_plot -s "' + perfmodel +'"')
		os.system('gnuplot starpu_'+perfmodel+'.gp')
		os.system('gv starpu_'+perfmodel+'.eps')

# acquire object
def acquire(obj, mode='R'):
	return starpupy.starpupy_acquire_object(obj, mode)

# release object
def release(obj):
	return starpupy.starpupy_release_object(obj)

# acquire object
def unregister(obj):
	return starpupy.starpupy_data_unregister_object(obj)

# acquire object
def unregister_submit(obj):
	return starpupy.starpupy_data_unregister_submit_object(obj)

def init():
        return starpupy.init()

def shutdown():
        return starpupy.shutdown()



# StarPU --- Runtime system for heterogeneous multicore architectures.
#
# Copyright (C) 2021, 2023  Université de Bordeaux, CNRS (LaBRI UMR 5800), Inria
#
# StarPU is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or (at
# your option) any later version.
#
# StarPU is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See the GNU Lesser General Public License in COPYING.LGPL for more details.
#
import starpu
from starpu import starpupy
from starpu import Handle
from starpu import HandleNumpy
try:
    import numpy as np
except (ModuleNotFoundError, ImportError):
	print("\n\nCan't find \"Python3 NumPy\" module (consider running \"pip3 install numpy\" or refer to https://numpy.org/install/)\n\n")
	starpupy.shutdown()
	exit(77)
import asyncio
import time
import array
import struct
try:
    import nest_asyncio
except ModuleNotFoundError as e:
	print("\n\nCan't find \"Python3 nest_asyncio\" module (consider running \"pip3 install nest_asyncio\")\n\n")
	starpupy.shutdown()
	exit(77)
import json
import sys
import statistics
import test_handle_bench

#############################perf test####################################
# Numpy function
@starpu.access(a="RW", b="R")
def add_numpy(a,b):
	np.add(a,b,out=a)

# custom function
@starpu.access(a="RW", b="R")
def add_custom(a,b):
    for i in range(np.size(a)):
        a[i] = a[i] + b[i]

program_submit1=[]
program_await1=[]

program_submit2=[]
program_await2=[]

program_submit3=[]

num=20
# calculate the standard deviasion
list_std11 = []
list_std12 = []
list_std21 = []
list_std22 = []
list_std3 = []

# using handle return future
def test_comp_handle_ret_fut(a,b):
	async def asy_main():
		start_exec1=0
		end_exec1=0
		start_exec2=0
		end_exec2=0
		list_submit=[]
		list_await=[]
		for t in range(num):
			#print("loop", t)
			start_exec1=time.time()
			res_fut = starpu.task_submit()(add_custom, a, b)
			end_exec1=time.time()

			list_submit.append(end_exec1-start_exec1)

			start_exec2=time.time()
			res = await res_fut
			end_exec2=time.time()

			list_await.append(end_exec2-start_exec2)

		program_submit1.append(statistics.mean(list_submit))
		program_await1.append(statistics.mean(list_await))

		list_std11.append(statistics.stdev(list_submit))
		list_std12.append(statistics.stdev(list_await))

	loop=asyncio.get_event_loop()
	nest_asyncio.apply()
	loop.run_until_complete(asy_main())

# using handle return handle
def test_comp_handle_ret_handle(a,b):
	async def asy_main():
		start_exec1=0
		end_exec1=0
		start_exec2=0
		end_exec2=0
		list_submit=[]
		list_await=[]
		for t in range(num):
			#print("loop", t)
			start_exec1=time.time()
			res_handle = starpu.task_submit(ret_handle=True)(add_custom, a, b)
			end_exec1=time.time()

			list_submit.append(end_exec1-start_exec1)

			start_exec2=time.time()
			starpupy.task_wait_for_all()
			end_exec2=time.time()

			list_await.append(end_exec2-start_exec2)

		program_submit2.append(statistics.mean(list_submit))
		program_await2.append(statistics.mean(list_await))

		list_std21.append(statistics.stdev(list_submit))
		list_std22.append(statistics.stdev(list_await))
	loop=asyncio.get_event_loop()
	nest_asyncio.apply()
	loop.run_until_complete(asy_main())

#without using starpu
def test_numpy(a,b):
	async def asy_main():
		start_exec1=0
		end_exec1=0
		list_submit=[]
		for t in range(num):
			start_exec1=time.time()
			add_numpy(a, b)
			end_exec1=time.time()

			list_submit.append(end_exec1-start_exec1)

		program_submit3.append(statistics.mean(list_submit))

		list_std3.append(statistics.stdev(list_submit))
	loop=asyncio.get_event_loop()
	nest_asyncio.apply()
	loop.run_until_complete(asy_main())


#with handle return future
for i in test_handle_bench.list_size:
	#print("i with handle return future is", i)
	A = np.arange(i)
	test_comp_handle_ret_fut(A, A)
	starpu.unregister(A)

#with handle return handle
for i in test_handle_bench.list_size:
	#print("i with handle return handle is", i)
	A = np.arange(i)
	test_comp_handle_ret_handle(A, A)
	starpu.unregister(A)

#without starpu
for i in test_handle_bench.list_size:
	A = np.arange(i)
	test_numpy(A, A)

retfut_dict={'program_submit':program_submit1, 'program_await': program_await1}
rethandle_dict={'program_submit':program_submit2, 'program_await': program_await2}
nostarpu_dict={'program_submit':program_submit3}

# print(retfut_dict)
# print(rethandle_dict)
# print(nostarpu_dict)

dict_std={'list_std11':list_std11, 'list_std12':list_std12, 'list_std21':list_std21, 'list_std22':list_std22, 'list_std3':list_std3}

#####write the dict in file#####
js1 = json.dumps(retfut_dict)
file1 = open('handle_perf1.txt', 'w')
file1.write(js1)
file1.close()

js2 = json.dumps(rethandle_dict)
file2 = open('handle_perf2.txt', 'w')
file2.write(js2)
file2.close()

js3 = json.dumps(nostarpu_dict)
file3 = open('handle_perf3.txt', 'w')
file3.write(js3)
file3.close()

js_std = json.dumps(dict_std)
file_std = open('handle_perf_std.txt', 'w')
file_std.write(js_std)
file_std.close()

starpupy.shutdown()



# StarPU --- Runtime system for heterogeneous multicore architectures.
#
# Copyright (C) 2020-2023  Universit'e de Bordeaux, CNRS (LaBRI UMR 5800), Inria
#
# StarPU is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or (at
# your option) any later version.
#
# StarPU is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See the GNU Lesser General Public License in COPYING.LGPL for more details.
#

try:
    import numpy as np
except (ModuleNotFoundError, ImportError):
	print("Can't find \"Python3 NumPy\" module (consider running \"pip3 install numpy\" or refer to https://numpy.org/install/)")
	exit(77)

import starpu
from starpu import Handle
from starpu import HandleNumpy
import asyncio
import time
import array
import struct

try:
        starpu.init()
except Exception as e:
        print(e)
        exit(77)

def await_fut(fut):
    return fut.result()

# 1-dimension
# arr = np.arange(20)

# 2-dimension
# n, m = 20, 10
# arr = np.arange(n*m).reshape(n, m)

# 3-dimension
# x, y, z = 10, 15, 20
# arr = np.arange(x*y*z).reshape(x, y, z)

# 4-dimension
x, y, z, t = 10, 5, 10, 20
arr = np.arange(x*y*z*t).reshape(x, y, z, t)
print("input array is", arr)

arr_h = Handle(arr)

# split into split_num of sub handles
split_num = 3
#arr_h_list = arr_h.partition(split_num, 0, [6,6,8])
arr_h_list = arr_h.partition(split_num, 0, [3,2,5])

n_arr = arr_h.get_partition_size(arr_h_list)

print("partition size is", n_arr)

def show(x):
    print("Function printing:", x)

@starpu.access(a="RW")
def add(a,b):
	np.add(a,b,out=a)

for i in range(split_num):
	starpu.task_submit(ret_handle=False,ret_fut=False)(add, arr_h_list[i], arr_h_list[i])

# async def main():
# 	for i in range(split_num):
# 		res=starpu.task_submit()(add, arr_h_list[i], arr_h_list[i])
# 		res1=await(res)
# asyncio.run(main())

arr_r = arr_h.acquire(mode='RW')
print("output array is:", arr_r)
arr_h.release()

arr_h.unpartition(arr_h_list, split_num)

arr_h.unregister()

starpu.shutdown()



#!/usr/bin/env python3
# coding=utf-8
#
# StarPU --- Runtime system for heterogeneous multicore architectures.
#
# Copyright (C) 2019-2023  Université de Bordeaux, CNRS (LaBRI UMR 5800), Inria
#
# StarPU is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or (at
# your option) any later version.
#
# StarPU is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See the GNU Lesser General Public License in COPYING.LGPL for more details.
#

"""
Offline tool to draw graph showing elapsed time between sent or received data and their use by tasks
"""

import sys
PROGNAME = sys.argv[0]

def usage():
    print("Offline tool to draw graph showing elapsed time between sent or received data and their use by tasks")
    print("")
    print("Usage: %s <folder containing comms.rec and tasks.rec files>" % PROGNAME)

if len(sys.argv) != 2:
    usage()
    sys.exit(1)

import re
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
import os

def convert_rec_file(filename):
    lines = []
    item = dict()

    with open(filename, "r") as f:
        for l in f.readlines():
            if l == "\n":
                lines.append(item)
                item = dict()
            else:
                ls = l.split(":")
                key = ls[0].lower()
                value = ls[1].strip()

                if key in item:
                    print("Warning: duplicated key '" + key + "'")
                else:
                    if re.match(r'^\d+$', value) != None:
                        item[key] = int(value)
                    elif re.match(r'^\d+\.\d+$', value) != None:
                        item[key] = float(value)
                    else:
                        item[key] = value

    return lines

working_directory = sys.argv[1]

comms = convert_rec_file(os.path.join(working_directory, "comms.rec"))
tasks = [t for t in
         convert_rec_file(os.path.join(working_directory, "tasks.rec")) if "control" not in t and "starttime" in t]

if len(tasks) == 0:
    print("There is no task using data after communication.")
    sys.exit(0)

def plot_graph(comm_time_key, match, filename, title, xlabel):
    workers = dict()
    durations = []
    min_time = 0.
    max_time = 0.

    for c in comms:
        t_matched = None
        for t in tasks:
            if match(t, c):
                t_matched = t
                break

        if t_matched is not None:
            worker = str(t_matched['mpirank']) + "-" + str(t_matched['workerid'])
            if worker not in workers:
                workers[worker] = []

            eps = t["starttime"] - c[comm_time_key]
            assert eps > 0
            durations.append(eps)
            workers[worker].append((c[comm_time_key], eps))

            if min_time == 0 or c[comm_time_key] < min_time:
                min_time = c[comm_time_key]
            if max_time == 0 or c[comm_time_key] > max_time:
                max_time = c[comm_time_key]

    fig = plt.figure(constrained_layout=True)

    gs = GridSpec(2, 2, figure=fig)
    axs = [fig.add_subplot(gs[0, :-1]), fig.add_subplot(gs[1, :-1]), fig.add_subplot(gs[0:, -1])]
    i = 0
    for y, x in workers.items():
        # print(y, x)
        axs[0].broken_barh(x, [i*10, 8], facecolors=(0.1, 0.2, 0.5, 0.2))
        i += 1

    i = 0
    for y, x in workers.items():
        for xx in x:
            axs[1].broken_barh([xx], [i, 1])
            i += 1

    axs[0].set_yticks([i*10+4 for i in range(len(workers))])
    axs[0].set_yticklabels(list(workers))
    axs[0].set(xlabel="Time (ms) - Duration: " + str(max_time - min_time) + "ms",
               ylabel="Worker [mpi]-[*pu]", title=title)

    if len(durations) != 0:
        axs[2].hist(durations, bins=np.logspace(np.log10(1), np.log10(max(durations)), 50), rwidth=0.8)
        axs[2].set_xscale("log")
        axs[2].set(xlabel=xlabel, ylabel="Number of occurrences", title="Histogramm")

    fig.set_size_inches(15, 9)

    plt.savefig(os.path.join(working_directory, filename), dpi=100)
    plt.show()

plot_graph("recvtime", lambda t,
           c: (t["mpirank"] == c["dst"] and t["starttime"] >= c["recvtime"] and str(c["recvhandle"]) in t["handles"]),
           "recv_use.png", "Elapsed time between recv and use (ms)", "Time between data reception and its use by a task")
plot_graph("sendtime", lambda t,
           c: (t["mpirank"] == c["src"] and t["starttime"] >= c["sendtime"] and str(c["sendhandle"]) in t["handles"]),
           "send_use.png", "Elapsed time between send and use (ms)", "Time between data sending and its use by a task")



#!/usr/bin/python3
# StarPU --- Runtime system for heterogeneous multicore architectures.
#
# Copyright (C) 2022-2023  Université de Bordeaux, CNRS (LaBRI UMR 5800), Inria
#
# StarPU is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or (at
# your option) any later version.
#
# StarPU is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See the GNU Lesser General Public License in COPYING.LGPL for more details.
#

import os
import operator
import sys

files = {}

with open(sys.argv[1]+"/doxygen-config.cfg", "r", encoding="utf-8") as fin:
    for line in fin.readlines():
        if ".doxy" in line and not "foreword.doxy" in line:
            for x in line.split(" "):
                xx = x.strip()
                if ".doxy" in xx:
                    with open(xx, "r", encoding="utf-8") as fin:
                        for line in fin.readlines():
                            if "\page" in line:
                                line = line.replace("/*! \page ", "").strip()
                                files[xx] = line[0:line.index(" ")]+".html"

htmlfiles = ["index.html"]
htmlfiles.extend(files.values())

htmldir=sys.argv[2]+"/"

chapter=0
for x in htmlfiles:
    chapter+=1
    section=0
    with open(htmldir+x, "r", encoding="utf-8") as fin:
        with open(htmldir+x+".count.html", "w", encoding="utf-8") as fout:
            for line in fin.readlines():
                if not "Foreword" in line:
                    if "<div class=\"title\">" in line:
                        line = line.replace("<div class=\"title\">", "<div class=\"title\">"+str(chapter)+". ")
                    if "<h1>" in line:
                        section += 1
                        line = line.replace("<h1>", "<h1>" + str(chapter) + "." + str(section))
                        subsection = 0
                    if "<h2>" in line:
                        subsection += 1
                        line = line.replace("<h2>", "<h2>" + str(chapter) + "." + str(section) + "." + str(subsection))
                        subsubsection = 0
                    if "<h3>" in line:
                        subsubsection += 1
                        line = line.replace("<h3>", "<h3>" + str(chapter) + "." + str(section) + "." + str(subsection) + "." + str(subsubsection))
                fout.write(line)
    os.rename(htmldir+x+".count.html", htmldir+x)



# StarPU --- Runtime system for heterogeneous multicore architectures.
#
# Copyright (C) 2020-2022  Universit'e de Bordeaux, CNRS (LaBRI UMR 5800), Inria
#
# StarPU is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or (at
# your option) any later version.
#
# StarPU is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See the GNU Lesser General Public License in COPYING.LGPL for more details.
#
from starpu import starpupy
import starpu
import asyncio
from functools import partial
import inspect

def delayed(f=None, **kwargs):
	# add options of task_submit
	if f is None:
		return partial(delayed, **kwargs)
	def submit(*args):
		# set the access right
		access_mode={}
		f_args = inspect.getfullargspec(f).args
		# check the access right of argument is set in mode or not
		for i in range(len(f_args)):
			if f_args[i] in kwargs.keys():
				# write access modes in f.access attribute
				access_mode[f_args[i]]=kwargs[f_args[i]]
				setattr(f, "starpu_access", access_mode)

		fut = starpu.task_submit(**kwargs)(f, *args)
		return fut
	return submit



#!/usr/bin/env python3
# coding=utf-8
#
# StarPU --- Runtime system for heterogeneous multicore architectures.
#
# Copyright (C) 2019-2023  Université de Bordeaux, CNRS (LaBRI UMR 5800), Inria
#
# StarPU is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or (at
# your option) any later version.
#
# StarPU is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See the GNU Lesser General Public License in COPYING.LGPL for more details.
#

"""
Plot statistics produced when running an application with STARPU_MPI_STATS=1
"""

import sys
import re
import os

PROGNAME=sys.argv[0]

def usage():
    print("Offline tool to draw a communication matrix")
    print("")
    print("Usage: %s <output_execution>" % PROGNAME)
    print("")
    print("Options:")
    print("	-h, --help          display this help and exit")
    print("	-v, --version       output version information and exit")
    print("	-png                produce plots in png format (default is pdf)")
    print("")
    print("Report bugs to <starpu-devel@inria.fr>")
    sys.exit(1)

if len(sys.argv) >= 2:
    if sys.argv[1] == '-v' or sys.argv[1] == '--version':
        print("%s (StarPU) 1.4.5" % PROGNAME)
        sys.exit(0)
    if sys.argv[1] == '-h' or sys.argv[1] == '--help':
        usage()
if len(sys.argv) == 1:
    usage()

if len(sys.argv) >= 2 and sys.argv[1] == '-png':
    outputformat='png'
    outputext='png'
    outputfile=sys.argv[2]
else:
    outputformat='pdf color'
    outputext='pdf'
    outputfile=sys.argv[1]

# find the number of nodes
nodes=0
file = open(outputfile, "r")
for line in file.readlines():
    if re.search('TOTAL', line):
        (node,stuff)=line.split(sep="[")[2].split("]")
        if int(node) > nodes:
            nodes=int(node)
file.close()
nodes=nodes+1

# extract volume of comm and bandwidth between all pair of nodes
volumes = [[0 for _ in range(nodes)] for _ in range(nodes)]
bandwidth = [[0 for _ in range(nodes)] for _ in range(nodes)]
file = open(outputfile, "r")
for line in file.readlines():
    if re.search(r'\[starpu_comm_stats]', line) and not re.search('TOTAL', line) and re.search('MB/s', line):
        (head,volB,B,volMB,MB,bwB,B,bwMB,MB) = line.split()
        (src,dst)=head.split(sep="[")[2].split(sep="]")[0].split(sep=":")
        volumes[int(src)][int(dst)] = float(volB)
        bandwidth[int(src)][int(dst)] = float(bwB)
file.close()

def write_data(filename, nodes, data):
    ofile=open(filename, "w")
    for dst in range(nodes):
        for src in range(nodes):
            ofile.write("%f "% data[src][dst])
        ofile.write("\n")
    ofile.close()

def generate_gnuplot_script(filename, datafilename, outputfile, nodes):
    ofile=open(filename, "w")
    srctics=""
    dsttics=""
    for node in range(nodes-1):
        srctics += "\"src%d\" %d, " % (node, node)
        dsttics += "\"dst%d\" %d, " % (node, node)
    ofile.write("set term %s\n" % outputformat)
    ofile.write("set output \"%s.%s\"\n" % (outputfile, outputext))
    ofile.write("set view map scale 1\nset style data lines\n")
    ofile.write("set palette gray\n")
    ofile.write("set xtics (%s\"src%d\" %d)\n" % (srctics, nodes-1, nodes-1))
    ofile.write("set ytics (%s\"dst%d\" %d)\n" % (dsttics, nodes-1, nodes-1))
    ofile.write("plot '%s' matrix with image\n" % datafilename)
    ofile.close()

# generate gnuplot volume data and script file
write_data(outputfile+"_volume.data", nodes, volumes)
generate_gnuplot_script(outputfile+"_volume.gp", outputfile+"_volume.data", outputfile+"_volume_heatmap", nodes)
os.system("gnuplot " + outputfile+"_volume.gp")
print("Generated file \"%s.%s\"" % (outputfile+"_volume.data", outputext))

# generate gnuplot bandwidth data and script file
write_data(outputfile+"_bw.data", nodes, bandwidth)
generate_gnuplot_script(outputfile+"_bw.gp", outputfile+"_bw.data", outputfile+"_bw_heatmap", nodes)
os.system("gnuplot " + outputfile+"_bw.gp")
print("Generated file \"%s.%s\"" % (outputfile+"_bw.data", outputext))



# StarPU --- Runtime system for heterogeneous multicore architectures.
#
# Copyright (C) 2021  Université de Bordeaux, CNRS (LaBRI UMR 5800), Inria
#
# StarPU is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or (at
# your option) any later version.
#
# StarPU is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See the GNU Lesser General Public License in COPYING.LGPL for more details.
#

listX = [10, 100, 1000, 10000, 100000, 1000000]
#listX = [10, 100]
list_size = []
for x in listX:
	for X in range(x, x*10, x):
		list_size.append(X)
list_size.append(10000000)
list_size.append(20000000)
list_size.append(30000000)
list_size.append(40000000)
list_size.append(50000000)
#print("list of size is",list_size)




#!/usr/bin/env python3
# coding=utf-8
#
# StarPU --- Runtime system for heterogeneous multicore architectures.
#
# Copyright (C) 2016-2023  Université de Bordeaux, CNRS (LaBRI UMR 5800), Inria
#
# StarPU is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or (at
# your option) any later version.
#
# StarPU is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See the GNU Lesser General Public License in COPYING.LGPL for more details.
#

"""
This script parses the generated trace.rec file and reports statistics about
the number of different events/tasks and their durations. The report is
similar to the starpu_paje_state_stats.in script, except that this one
doesn't need R and pj_dump (from the pajeng repository), and it is also much
faster.
"""

import getopt
import os
import sys

class Event(object):
    def __init__(self, type, name, category, start_time):
        self._type = type
        self._name = name
        self._category = category
        self._start_time = start_time

class EventStats(object):
    def __init__(self, name, duration_time, category, count = 1):
        self._name = name
        self._duration_time = duration_time
        self._category = category
        self._count = count

    def aggregate(self, duration_time):
        self._duration_time += duration_time
        self._count += 1

    def show(self):
        if not self._name == None and not self._category == None:
            print("\"" + self._name + "\"," + str(self._count) + ",\"" +
                  self._category + "\"," + str(round(self._duration_time, 6)))

class Worker(object):
    def __init__(self, id):
        self._id        = id
        self._events    = []
        self._stats     = []
        self._stack     = []
        self._current_state = None

    def get_event_stats(self, name):
        for stat in self._stats:
            if stat._name == name:
                return stat
        return None

    def add_event(self, type, name, category, start_time):
        self._events.append(Event(type, name, category, start_time))

    def add_event_to_stats(self, curr_event):
        if curr_event._type == "PushState":
            self._stack.append(curr_event)
            # Will look later to find a PopState event.
            return
        elif curr_event._type == "PopState":
            if len(self._stack) == 0:
                print("warning: PopState without a PushState, probably a trace with start/stop profiling")
                self._current_state = None
                return
            next_event = curr_event
            curr_event = self._stack.pop()
        elif curr_event._type == "SetState":
            if self._current_state == None:
                # First SetState event found
                self._current_state = curr_event
                return
            saved_state = curr_event
            next_event = curr_event
            curr_event = self._current_state
            self._current_state = saved_state
        else:
            sys.exit("ERROR: Invalid event type!")

        # Compute duration with the next event.
        a = curr_event._start_time
        b = next_event._start_time

        # Add the event to the list of stats.
        for i in range(len(self._stats)):
            if self._stats[i]._name == curr_event._name:
                self._stats[i].aggregate(b - a)
                return
        self._stats.append(EventStats(curr_event._name, b - a,
                                      curr_event._category))

    def calc_stats(self, start_profiling_times, stop_profiling_times):
        num_events = len(self._events)
        use_start_stop = len(start_profiling_times) != 0
        for i in range(0, num_events):
            event = self._events[i]
            if i > 0 and self._events[i-1]._name == "Deinitializing":
                # Drop all events after the Deinitializing event is found
                # because they do not make sense.
                break

            if not use_start_stop:
                self.add_event_to_stats(event)
                continue

            # Check if the event is in between start/stop profiling events
            for t in range(len(start_profiling_times)):
                if (event._start_time > start_profiling_times[t] and
                    event._start_time < stop_profiling_times[t]):
                    self.add_event_to_stats(event)
                    break

        if not use_start_stop:
            return

        # Special case for SetState events which need a next one for computing
        # the duration.
        curr_event = self._events[-1]
        if curr_event._type == "SetState":
            for i in range(len(start_profiling_times)):
                if (curr_event._start_time > start_profiling_times[i] and
                    curr_event._start_time < stop_profiling_times[i]):
                    curr_event = Event(curr_event._type, curr_event._name,
                                       curr_event._category,
                                       stop_profiling_times[i])
            self.add_event_to_stats(curr_event)

def read_blocks(input_file):
    empty_lines = 0
    first_line = 1
    blocks = []
    for line in open(input_file):
        if first_line:
            blocks.append([])
            blocks[-1].append(line)
            first_line = 0

        # Check for empty lines
        if not line or line[0] == '\n':
            # If 1st one: new block
            if empty_lines == 0:
                blocks.append([])
            empty_lines += 1
        else:
            # Non empty line: add line in current(last) block
            empty_lines = 0
            blocks[-1].append(line)
    return blocks

def read_field(field, index):
    return field[index+1:-1]

def insert_worker_event(workers, prog_events, block):
    worker_id = -1
    name = None
    start_time = 0.0
    category = None

    for line in block:
        key   = line[:2]
        value = read_field(line, 2)
        if key == "E:": # EventType
            event_type = value
        elif key == "C:": # Category
            category = value
        elif key == "W:": # WorkerId
            worker_id = int(value)
        elif key == "N:": # Name
            name = value
        elif key == "S:": # StartTime
            start_time = float(value)

    # Program events don't belong to workers, they are globals.
    if category == "Program":
        prog_events.append(Event(event_type, name, category, start_time))
        return

    for worker in workers:
        if worker._id == worker_id:
            worker.add_event(event_type, name, category, start_time)
            return
    worker = Worker(worker_id)
    worker.add_event(event_type, name, category, start_time)
    workers.append(worker)

def calc_times(stats):
    tr = 0.0 # Runtime
    tt = 0.0 # Task
    ti = 0.0 # Idle
    ts = 0.0 # Scheduling
    for stat in stats:
        if stat._category == None:
            continue
        if stat._category == "Runtime":
            if stat._name == "Scheduling":
                # Scheduling time is part of runtime but we want to have
                # it separately.
                ts += stat._duration_time
            else:
                tr += stat._duration_time
        elif stat._category == "Task":
            tt += stat._duration_time
        elif stat._category == "Other":
            ti += stat._duration_time
        else:
            print("WARNING: Unknown category '" + stat._category + "'!")
    return ti, tr, tt, ts

def save_times(ti, tr, tt, ts):
    f = open("times.csv", "w+")
    f.write("\"Time\",\"Duration\"\n")
    f.write("\"Runtime\"," + str(tr) + "\n")
    f.write("\"Task\"," + str(tt) + "\n")
    f.write("\"Idle\"," + str(ti) + "\n")
    f.write("\"Scheduling\"," + str(ts) + "\n")
    f.close()

def calc_et(tt_1, tt_p):
    """ Compute the task efficiency (et). This measures the exploitation of
    data locality. """
    return tt_1 / tt_p

def calc_es(tt_p, ts_p):
    """ Compute the scheduling efficiency (es). This measures time spent in
    the runtime scheduler. """
    return tt_p / (tt_p + ts_p)

def calc_er(tt_p, tr_p, ts_p):
    """ Compute the runtime efficiency (er). This measures how the runtime
    overhead affects performance."""
    return (tt_p + ts_p) / (tt_p + tr_p + ts_p)

def calc_ep(tt_p, tr_p, ti_p, ts_p):
    """ Compute the pipeline efficiency (et). This measures how much
    concurrency is available and how well it's exploited. """
    return (tt_p + tr_p + ts_p) / (tt_p + tr_p + ti_p + ts_p)

def calc_e(et, er, ep, es):
    """ Compute the parallel efficiency. """
    return et * er * ep * es

def save_efficiencies(e, ep, er, et, es):
    f = open("efficiencies.csv", "w+")
    f.write("\"Efficiency\",\"Value\"\n")
    f.write("\"Parallel\"," + str(e) + "\n")
    f.write("\"Task\"," + str(et) + "\n")
    f.write("\"Runtime\"," + str(er) + "\n")
    f.write("\"Scheduling\"," + str(es) + "\n")
    f.write("\"Pipeline\"," + str(ep) + "\n")
    f.close()

def usage():
    print("USAGE:")
    print("starpu_trace_state_stats.py [ -te -s=<time> ] <trace.rec>")
    print("")
    print("OPTIONS:")
    print(" -t or --time            Compute and dump times to times.csv")
    print("")
    print(" -e or --efficiency      Compute and dump efficiencies to efficiencies.csv")
    print("")
    print(" -s or --seq_task_time   Used to compute task efficiency between sequential and parallel times")
    print("                         (if not set, task efficiency will be 1.0)")
    print("")
    print("EXAMPLES:")
    print("# Compute event statistics and report them to stdout:")
    print("starpu_trace_state_stats.py trace.rec")
    print("")
    print("# Compute event stats, times and efficiencies:")
    print("starpu_trace_state_stats.py -te trace.rec")
    print("")
    print("# Compute correct task efficiency with the sequential task time:")
    print("starpu_trace_state_stats.py -s=60093.950614 trace.rec")

def main():
    try:
        opts, args = getopt.getopt(sys.argv[1:], "hets:",
                                   ["help", "time", "efficiency", "seq_task_time="])
    except getopt.GetoptError:
        usage()
        sys.exit(1)

    dump_time = False
    dump_efficiency = False
    tt_1 = 0.0

    for o, a in opts:
        if o in ("-h", "--help"):
            usage()
            sys.exit()
        elif o in ("-t", "--time"):
            dump_time = True
        elif o in ("-e", "--efficiency"):
            dump_efficiency = True
        elif o in ("-s", "--seq_task_time"):
            tt_1 = float(a)

    if len(args) < 1:
        usage()
        sys.exit()
    recfile = args[0]

    if not os.path.isfile(recfile):
        sys.exit("File does not exist!")

    # Declare a list for all workers.
    workers = []

    # Declare a list for program events
    prog_events = []

    # Read the recutils file format per blocks.
    blocks = read_blocks(recfile)
    for block in blocks:
        if not len(block) == 0:
            first_line = block[0]
            if first_line[:2] == "E:":
                insert_worker_event(workers, prog_events, block)

    # Find allowed range times between start/stop profiling events.
    start_profiling_times = []
    stop_profiling_times = []
    for prog_event in prog_events:
        if prog_event._name == "start_profiling":
            start_profiling_times.append(prog_event._start_time)
        if prog_event._name == "stop_profiling":
            stop_profiling_times.append(prog_event._start_time)

    if len(start_profiling_times) != len(stop_profiling_times):
        sys.exit("Mismatch number of start/stop profiling events!")

    # Compute worker statistics.
    stats = []
    for worker in workers:
        worker.calc_stats(start_profiling_times, stop_profiling_times)
        for stat in worker._stats:
            found = False
            for s in stats:
                if stat._name == s._name:
                    found = True
                    break
            if not found == True:
                stats.append(EventStats(stat._name, 0.0, stat._category, 0))

    # Compute global statistics for all workers.
    for i in range(0, len(workers)):
        for stat in stats:
            s = workers[i].get_event_stats(stat._name)
            if not s == None:
                # A task might not be executed on all workers.
                stat._duration_time += s._duration_time
                stat._count += s._count

    # Output statistics.
    print("\"Name\",\"Count\",\"Type\",\"Duration\"")
    for stat in stats:
        stat.show()

    # Compute runtime, task, idle, scheduling times and dump them to times.csv
    ti_p = tr_p = tt_p = ts_p = 0.0
    if dump_time == True:
        ti_p, tr_p, tt_p, ts_p = calc_times(stats)
        save_times(ti_p, tr_p, tt_p, ts_p)

    # Compute runtime, task, idle efficiencies and dump them to
    # efficiencies.csv.
    if dump_efficiency == True or not tt_1 == 0.0:
        if dump_time == False:
            ti_p, tr_p, tt_p, ts_p = calc_times(stats)
        if tt_1 == 0.0:
            sys.stderr.write("WARNING: Task efficiency will be 1.0 because -s is not set!\n")
            tt_1 = tt_p

        # Compute efficiencies.
        et = round(calc_et(tt_1, tt_p), 6)
        es = round(calc_es(tt_p, ts_p), 6)
        er = round(calc_er(tt_p, tr_p, ts_p), 6)
        ep = round(calc_ep(tt_p, tr_p, ti_p, ts_p), 6)
        e  = round(calc_e(et, er, ep, es), 6)
        save_efficiencies(e, ep, er, et, es)
if __name__ == "__main__":
    main()



# StarPU --- Runtime system for heterogeneous multicore architectures.
#
# Copyright (C) 2020-2023  Universit'e de Bordeaux, CNRS (LaBRI UMR 5800), Inria
#
# StarPU is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or (at
# your option) any later version.
#
# StarPU is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See the GNU Lesser General Public License in COPYING.LGPL for more details.
#

import sys
import types
import joblib as jl
from joblib import logger
from joblib._parallel_backends import ParallelBackendBase
from starpu import starpupy
from starpu import Handle
import starpu
import asyncio
import math
import functools
try:
    import numpy as np
    has_numpy=True
except:
    has_numpy=False
import inspect
import threading

loop = asyncio.get_event_loop()
if (loop.is_running()):
	try:
		import nest_asyncio
		nest_asyncio.apply()
		has_nest=True
	except (ModuleNotFoundError, ImportError):
		has_nest=False

BACKENDS={
	#'loky': LokyBackend,
}
_backend = threading.local()

# get the number of CPUs controlled by StarPU
def cpu_count():
	n_cpus=starpupy.worker_get_count_by_type(starpu.STARPU_CPU_WORKER)
	return n_cpus

# split a list ls into n_block numbers of sub-lists
def partition(ls, n_block):
	if len(ls)>=n_block:
		# there are n1 sub-lists which contain q1 elements, and (n_block-n1) sublists which contain q2 elements (n1 can be 0)
		q1=math.ceil(len(ls)/n_block)
		q2=math.floor(len(ls)/n_block)
		n1=len(ls)%n_block
		#n2=n_block-n1
		# generate n1 sub-lists in L1, and (n_block-n1) sub-lists in L2
		L1=[ls[i:i+q1] for i in range(0, n1*q1, q1)]
		L2=[ls[i:i+q2] for i in range(n1*q1, len(ls), q2)]

		L=L1+L2
	else:
		# if the block number is larger than the length of list, each element in the list is a sub-list
		L=[ls[i:i+1] for i in range (len(ls))]
	return L

# split a two-dimension numpy matrix into n_block numbers of sub-matrices
def array2d_split(a, n_block):
	# decompose number of n_jobs to two integers multiply
	c_tmp=math.floor(math.sqrt(n_block))
	for i in range (c_tmp,0,-1):
		if n_block%i==0:
			c=i
			r=int(n_block/c)
			break
	# split column
	arr_split_c=np.array_split(a,c,0)
	arr_split=[]
	# split row
	for i in range(c):
		arr_split_r=np.array_split(arr_split_c[i],r,1)
		for j in range(r):
			arr_split.append(arr_split_r[j].copy(order='C'))
	return arr_split


def future_generator(iterable, n_jobs, dict_task):
	# iterable is generated by delayed function, after converting to a list, the format is [function, (arg1, arg2, ... ,)]
	#print("iterable type is ", type(iterable))
	#print("iterable is", iterable)
	# get the number of block
	if n_jobs<-cpu_count()-1:
		raise SystemExit('Error: n_jobs is out of range, number of CPUs is', cpu_count())
	elif n_jobs<0:
		n_block=cpu_count()+1+n_jobs
	else:
		n_block=n_jobs
	if (n_block <= 0):
		n_block = 1

	# if arguments is tuple format
	if type(iterable) is tuple:
		# the function is always the first element
		f=iterable[0]
		# get the name of formal arguments of f
		formal_args=inspect.getfullargspec(f).args
		# get the arguments list
		args=[]
		# argument is arbitrary in iterable[1]
		args=list(iterable[1])
		# argument is keyword argument in iterable[2]
		for i in range(len(formal_args)):
			for j in iterable[2].keys():
				if j==formal_args[i]:
					args.append(iterable[2][j])
		# check whether all arrays have the same size
		l_arr=[]
		# list of Future result
		L_fut=[]
		# split the vector
		args_split=[]
		# handle list
		arg_h=[]
		for i in range(len(args)):
			args_split.append([])
			# if the array is an numpy array
			if has_numpy and type(args[i]) is np.ndarray:
				# check whether the arg is already registered
				handle_dict = starpu.handle_dict
				if handle_dict.get(id(args[i]))==None:
					arr_h = Handle(args[i])
					arg_h.append(arr_h)
					args_split[i] = arr_h.partition(n_block, 0)
				else:
					arr_h = handle_dict.get(id(args[i]))
					arg_h.append(arr_h)
					args_split[i] = arr_h.partition(n_block, 0)
			# if the array is a generator
			elif isinstance(args[i],types.GeneratorType):
				# split generator
				args_split[i]=partition(list(args[i]),n_block)
				arg_h.append(None)
				# get the length of generator
				l_arr.append(sum(len(args_split[i][j]) for j in range(len(args_split[i]))))
			else:
				arg_h.append(None)
		if len(set(l_arr))>1:
			raise SystemExit('Error: all arrays should have the same size')
		#print("args list is", args_split)
		for i in range(n_block):
			# generate the argument list
			L_args=[]
			sizebase=0
			for j in range(len(args)):
				if (has_numpy and type(args[j]) is np.ndarray):
					L_args.append(args_split[j][i])
					n_arr = arg_h[j].get_partition_size(args_split[j])
					if sizebase==0:
						sizebase=n_arr[i]
					elif sizebase==n_arr[i]:
						continue
					else:
						raise SystemExit('Error: all arrays should be split into equal size')
				elif isinstance(args[j],types.GeneratorType):
					L_args.append(args_split[j][i])
					if sizebase==0:
						sizebase=len(args_split[j][i])
					elif sizebase==len(args_split[j][i]):
						continue
					else:
						raise SystemExit('Error: all arrays should be split into equal size')
				else:
					L_args.append(args[j])
			#print("L_args is", L_args)
			fut=starpu.task_submit(name=dict_task['name'], synchronous=dict_task['synchronous'], priority=dict_task['priority'],\
								   color=dict_task['color'], flops=dict_task['flops'], perfmodel=dict_task['perfmodel'], sizebase=sizebase,\
								   ret_handle=dict_task['ret_handle'], ret_fut=dict_task['ret_fut'], arg_handle=dict_task['arg_handle'], modes=dict_task['modes'])\
				                  (f, *L_args)
			L_fut.append(fut)
		# unpartition and unregister the numpy array
		for i in range(len(args)):
			if (has_numpy and type(args[i]) is np.ndarray):
				arg_h[i].unpartition(args_split[i], n_block)
				arg_h[i].unregister()
		return L_fut

	# if iterable is a generator or a list of function
	else:
		L=list(iterable)
		#print(L)
		# generate a list of function according to iterable
		def lf(ls):
			L_func=[]
			for i in range(len(ls)):
				# the first element is the function
				f=ls[i][0]
				# the second element is the args list of a type tuple
				L_args=list(ls[i][1])
				# generate a list of function
				L_func.append(f(*L_args))
			return L_func

		# generate the split function list
		L_split=partition(L,n_block)
		# operation in each split list
		L_fut=[]
		for i in range(len(L_split)):
			sizebase=len(L_split[i])
			fut=starpu.task_submit(name=dict_task['name'], synchronous=dict_task['synchronous'], priority=dict_task['priority'],\
								   color=dict_task['color'], flops=dict_task['flops'], perfmodel=dict_task['perfmodel'], sizebase=sizebase,\
								   ret_handle=dict_task['ret_handle'], ret_fut=dict_task['ret_fut'], arg_handle=dict_task['arg_handle'], modes=dict_task['modes'])\
				                  (lf, L_split[i])
			L_fut.append(fut)
		return L_fut

class Parallel(object):
	def __init__(self, mode="normal", perfmodel=None, end_msg=None,\
			 name=None, synchronous=0, priority=0, color=None, flops=None,\
			 ret_handle=False, ret_fut=True, arg_handle=True, modes=None,\
	         n_jobs=None, backend=None, verbose=0, timeout=None, pre_dispatch='2 * n_jobs',\
	         batch_size='auto', temp_folder=None, max_nbytes='1M',\
	         mmap_mode='r', prefer=None, require=None):
		#active_backend= get_active_backend()
		# nesting_level = active_backend.nesting_level

		# if backend is None:
		# 	backend = active_backend

		# else:
		# 	try:
		# 		backend_factory = BACKENDS[backend]
		# 	except KeyError as e:
		# 		raise ValueError("Invalid backend: %s, expected one of %r"
  #                                % (backend, sorted(BACKENDS.keys()))) from e
		# 	backend = backend_factory(nesting_level=nesting_level)

		if n_jobs is None:
			n_jobs = 1

		self.mode=mode
		self.perfmodel=perfmodel
		self.end_msg=end_msg
		self.name=name
		self.synchronous=synchronous
		self.priority=priority
		self.color=color
		self.flops=flops
		self.ret_handle=ret_handle
		self.ret_fut=ret_fut
		self.arg_handle=arg_handle
		self.modes=modes
		self.n_jobs=n_jobs
		self._backend=backend

	def print_progress(self):
		#todo
		print("", starpupy.task_nsubmitted())

	def __call__(self,iterable):
		#generate the dictionary of task_submit
		dict_task={'name': self.name, 'synchronous': self.synchronous, 'priority': self.priority, 'color': self.color, 'flops': self.flops, 'perfmodel': self.perfmodel, 'ret_handle': self.ret_handle, 'ret_fut': self.ret_fut, 'arg_handle': self.arg_handle, 'modes': self.modes}
		if hasattr(self._backend, 'start_call'):
			self._backend.start_call()
		# the mode normal, user can call the function directly without using async
		if self.mode=="normal":
			async def asy_main():
				L_fut=future_generator(iterable, self.n_jobs, dict_task)
				res=[]
				for i in range(len(L_fut)):
					L_res=await L_fut[i]
					if L_res is None:
						res=None
					else:
						res.extend(L_res)
				#print(res)
				#print("type of result is", type(res))
				return res
			#asyncio.run(asy_main())
			#retVal=asy_main
			#loop = asyncio.get_event_loop()
			if(loop.is_running() and not has_nest):
				raise starpupy.error("Can't find \'nest_asyncio\' module (consider running \"pip3 install nest_asyncio\" or try to remove \"-m asyncio\" when starting Python interpreter)")

			results = loop.run_until_complete(asy_main())
			retVal = results
		# the mode future, user needs to use asyncio module and await the Future result in main function
		elif self.mode=="future":
			L_fut=future_generator(iterable, self.n_jobs, dict_task)
			fut=asyncio.gather(*L_fut)
			if self.end_msg!=None:
				fut.add_done_callback(functools.partial(print, self.end_msg))
			retVal=fut
		if hasattr(self._backend, 'stop_call'):
			self._backend.stop_call()
		return retVal

def delayed(function):
	def delayed_function(*args, **kwargs):
		return function, args, kwargs
	return delayed_function


######################################################################
__version__ = jl.__version__

class Memory(jl.Memory):
	def __init__(self,location=None, backend='local', cachedir=None,
                 mmap_mode=None, compress=False, verbose=1, bytes_limit=None,
                 backend_options=None):
		super(Memory, self).__init__(location=None, backend='local', cachedir=None,
                 mmap_mode=None, compress=False, verbose=1, bytes_limit=None,
                 backend_options=None)


def dump(value, filename, compress=0, protocol=None, cache_size=None):
	return jl.dump(value, filename, compress, protocol, cache_size)

def load(filename, mmap_mode=None):
	return jl.load(filename, mmap_mode)

def hash(obj, hash_name='md5', coerce_mmap=False):
	return jl.hash(obj, hash_name, coerce_mmap)

def register_compressor(compressor_name, compressor, force=False):
	return jl.register_compressor(compressor_name, compressor, force)

def effective_n_jobs(n_jobs=-1):
	return cpu_count()

def get_active_backend():
	backend_and_jobs = getattr(_backend, 'backend_and_jobs', None)
	if backend_and_jobs is not None:
		backend,n_jobs=backend_and_jobs
		return backend
	backend = BACKENDS[loky](nesting_level=0)
	return backend

class parallel_backend(object):
	def __init__(self, backend, n_jobs=-1, inner_max_num_threads=None,
                 **backend_params):
		if isinstance(backend, str):
			backend = BACKENDS[backend](**backend_params)

		current_backend_and_jobs = getattr(_backend, 'backend_and_jobs', None)
		if backend.nesting_level is None:
			if current_backend_and_jobs is None:
				nesting_level = 0
			else:
				nesting_level = current_backend_and_jobs[0].nesting_level

			backend.nesting_level = nesting_level

		# Save the backends info and set the active backend
		self.old_backend_and_jobs = current_backend_and_jobs
		self.new_backend_and_jobs = (backend, n_jobs)

		_backend.backend_and_jobs = (backend, n_jobs)

	def __enter__(self):
		return self.new_backend_and_jobs

	def __exit__(self, type, value, traceback):
		self.unregister()

	def unregister(self):
		if self.old_backend_and_jobs is None:
			if getattr(_backend, 'backend_and_jobs', None) is not None:
				del _backend.backend_and_jobs
		else:
			_backend.backend_and_jobs = self.old_backend_and_jobs

def register_parallel_backend(name, factory):
	BACKENDS[name] = factory



# StarPU --- Runtime system for heterogeneous multicore architectures.
#
# Copyright (C) 2021, 2023  Université de Bordeaux, CNRS (LaBRI UMR 5800), Inria
#
# StarPU is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or (at
# your option) any later version.
#
# StarPU is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See the GNU Lesser General Public License in COPYING.LGPL for more details.
#
import starpu
from starpu import starpupy
from starpu import Handle
from starpu import HandleNumpy
try:
    import numpy as np
except (ModuleNotFoundError, ImportError):
	print("Can't find \"Python3 NumPy\" module (consider running \"pip3 install numpy\" or refer to https://numpy.org/install/)")
	starpupy.shutdown()
	exit(77)
import asyncio
import time
import array
import struct
import nest_asyncio
import json
import sys
import statistics

#############################perf test####################################
# Numpy function
@starpu.access(a="RW", b="R")
def add(a,b):
	np.add(a,b,out=a)

# custom function
# @starpu.access(a="RW", b="R")
# def add(a,b):
# 	for i in range(np.size(a)):
# 		a[i] = a[i] + b[i]

listX = [10, 100, 1000, 10000, 100000, 1000000]
list_size = []
for x in listX:
	for X in range(x, x*10, x):
		list_size.append(X)
list_size.append(10000000)
list_size.append(20000000)
list_size.append(30000000)
list_size.append(40000000)
list_size.append(50000000)
#print("list of size is",list_size)

program_submit1=[]
program_await1=[]

program_submit2=[]
program_await2=[]

program_submit3=[]

num=20
# calculate the standard deviasion
list_std11 = []
list_std12 = []
list_std21 = []
list_std22 = []
list_std3 = []

# using handle
def test_comp_handle(a,b):
	async def asy_main():
		start_exec1=0
		end_exec1=0
		start_exec2=0
		end_exec2=0
		list_submit=[]
		list_await=[]
		for t in range(num):
			#print("loop", t)
			start_exec1=time.time()
			res_fut = starpu.task_submit()(add, a, b)
			end_exec1=time.time()

			list_submit.append(end_exec1-start_exec1)

			start_exec2=time.time()
			res = await res_fut
			end_exec2=time.time()

			list_await.append(end_exec2-start_exec2)
		
		program_submit1.append(statistics.mean(list_submit))
		program_await1.append(statistics.mean(list_await))

		list_std11.append(statistics.stdev(list_submit))
		list_std12.append(statistics.stdev(list_await))

	loop=asyncio.get_event_loop()
	nest_asyncio.apply()
	loop.run_until_complete(asy_main())

#without using handle
def test_comp(a,b):
	async def asy_main():
		start_exec1=0
		end_exec1=0
		start_exec2=0
		end_exec2=0
		list_submit=[]
		list_await=[]
		for t in range(num):
			#print("loop", t)
			start_exec1=time.time()
			res_fut = starpu.task_submit(arg_handle=False)(add, a, b)
			end_exec1=time.time()

			list_submit.append(end_exec1-start_exec1)

			start_exec2=time.time()
			res = await res_fut
			end_exec2=time.time()

			list_await.append(end_exec2-start_exec2)

		program_submit2.append(statistics.mean(list_submit))
		program_await2.append(statistics.mean(list_await))

		list_std21.append(statistics.stdev(list_submit))
		list_std22.append(statistics.stdev(list_await))
		
	loop=asyncio.get_event_loop()
	nest_asyncio.apply()
	loop.run_until_complete(asy_main())
	
#without using starpu
def test_numpy(a,b):
	async def asy_main():
		start_exec1=0
		end_exec1=0
		list_submit=[]
		for t in range(num):
			start_exec1=time.time()
			add(a, b)
			end_exec1=time.time()
			
			list_submit.append(end_exec1-start_exec1)

		program_submit3.append(statistics.mean(list_submit))

		list_std3.append(statistics.stdev(list_submit))

	loop=asyncio.get_event_loop()
	nest_asyncio.apply()
	loop.run_until_complete(asy_main())


#with handle
for i in list_size:
	#print("i with handle is", i)
	A = np.arange(i)
	test_comp_handle(A, A)

	starpu.unregister(A)

#without handle	
for i in list_size:
	#print("i without handle is", i)
	A = np.arange(i)
	test_comp(A, A)

#without starpu
for i in list_size:
	A = np.arange(i)
	test_numpy(A, A)


withhandle_dict={'program_submit':program_submit1, 'program_await': program_await1}
nohandle_dict={'program_submit':program_submit2, 'program_await': program_await2}
nostarpu_dict={'program_submit':program_submit3}

# print(withhandle_dict)
# print(nohandle_dict)
# print(nostarpu_dict)

dict_std={'list_std11':list_std11, 'list_std12':list_std12, 'list_std21':list_std21, 'list_std22':list_std22, 'list_std3':list_std3}

#####write the dict in file#####
js1 = json.dumps(withhandle_dict)   
file1 = open('handle_perf1.txt', 'w')  
file1.write(js1)  
file1.close()

js2 = json.dumps(nohandle_dict)   
file2 = open('handle_perf2.txt', 'w')  
file2.write(js2)  
file2.close() 

js3 = json.dumps(nostarpu_dict)   
file3 = open('handle_perf3.txt', 'w')  
file3.write(js3)  
file3.close() 

js_std = json.dumps(dict_std)
file_std = open('handle_perf_std.txt', 'w')
file_std.write(js_std)
file_std.close()

starpupy.shutdown()



# StarPU --- Runtime system for heterogeneous multicore architectures.
#
# Copyright (C) 2021-2023  Université de Bordeaux, CNRS (LaBRI UMR 5800), Inria
#
# StarPU is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or (at
# your option) any later version.
#
# StarPU is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See the GNU Lesser General Public License in COPYING.LGPL for more details.
#
import starpu
from starpu import starpupy

import time
import sys
import getopt
import asyncio
import cProfile
import sys

mincpus = 1
maxcpus = starpupy.worker_get_count_by_type(starpu.STARPU_CPU_WORKER)
cpustep = 1

mintime = 128
maxtime = 128*1024
factortime = 2

ntasks = 64
nbuffers = 0
total_nbuffers = 0

#################parameters##############
try:
	opts, args = getopt.getopt(sys.argv[1:],"i:b:B:c:C:s:t:T:f:h")
except getopt.GetoptError:
	print("Usage:", sys.argv[0], "\n"\
	 "\t[-h help] \n "\
	 "\t[-i ntasks] [-b nbuffers] [-B total_nbuffers] \n"\
	 "\t[-c mincpus] [ -C maxcpus] [-s cpustep]\n"\
	 "\t[-t mintime] [-T maxtime] [-f factortime]")
	starpupy.shutdown()
	sys.exit(1)
for opt, arg in opts:
	if opt == '-i':
		ntasks = int(arg)
	elif opt == '-b':
		nbuffers = int(arg)
	elif opt == '-B':
		total_nbuffers = int(arg)
	elif opt == '-c':
		mincpus = int(arg)
	elif opt == '-C':
		maxcpus = int(arg)
	elif opt == '-s':
		cpustep = int(arg)
	elif opt == '-t':
		mintime = int(arg)
	elif opt == '-T':
		maxtime = int(arg)
	elif opt == '-f':
		factortime = int(arg)
	elif opt == '-h':
		print("Usage:", sys.argv[0], "[-h help] \n "\
		 "\t[-i ntasks] [-b nbuffers] [-B total_nbuffers] \n"\
		 "\t[-c mincpus] [ -C maxcpus] [-s cpustep]\n"\
		 "\t[-t mintime] [-T maxtime] [-f factortime]\n")
		print("runs \'ntasks\' tasks\n"\
		"- using \'nbuffers\' data each, randomly among \'total_nbuffers\' choices,\n"\
		"- with varying task durations, from \'mintime\' to \'maxtime\' (using \'factortime\')\n"\
		"- on varying numbers of cpus, from \'mincpus\' to \'maxcpus\' (using \'cpustep\')\n"\
		"\n"\
		"currently selected parameters: ", ntasks, " tasks using ", nbuffers, " buffers among ", total_nbuffers, \
		", from ", mintime, "us to ", maxtime, "us (factor ", factortime, "), from ", mincpus, " cpus to ", maxcpus, " cpus (step ", cpustep, ")", sep='')
		starpupy.shutdown()
		sys.exit(0)

########################################

# multiplication increment
def range_multi(start, end, factor):
	val_multi = []
	val = start
	while val <= end:
		val_multi.append(val)
		val = val * factor 
	return val_multi

# the test function
def func_test(t):
	time.sleep(t/1000000)

#pr = cProfile.Profile()

f = open("tasks_size_overhead.output",'w')

method="handle"
if len(sys.argv) > 1:
        method=sys.argv[1]

print("# tasks :", ntasks, "buffers :", nbuffers, "totoal_nbuffers :", total_nbuffers, file=f)
print("# ncups", end='\t', file=f)
for size in range_multi(mintime, maxtime, factortime):
	print(size, "iters(us)\ttotal(s)", end='\t', file=f)
print(end='\n', file=f)

print("\"seq\"\t", end=' ', file=f)
for size in range_multi(mintime, maxtime, factortime):
	#print("time size is", size)
	dstart=time.time()
	for i in range(ntasks):
		func_test(size)
	dend=time.time()
	print(int((dend-dstart)/ntasks*1000000), "\t", dend-dstart, end='\t', file=f)
	#print(size, "\t", dend-dstart, end='\t', file=f)
print(end='\n', file=f)

#pr.enable()

if method == "handle":
        # return value is handle
        for ncpus in range(mincpus, maxcpus+1, cpustep):
                starpupy.set_ncpu(ncpus)
                #print("ncpus is", ncpus)
                print(ncpus, end='\t', file=f)
                for size in range_multi(mintime, maxtime, factortime):
                        #print("time size is", size)
                        start=time.time()
                        for i in range(ntasks*ncpus):
                                res=starpu.task_submit(ret_handle=True)("func_test", size)
                        starpupy.task_wait_for_all()
                        end=time.time()
                        timing = end-start
                        print(size, "\t", timing/ncpus, end='\t', file=f)
                print(end='\n', file=f)

elif method == "futur":
        # return value is future
        async def main():
                for ncpus in range(mincpus, maxcpus+1, cpustep):
                        starpupy.set_ncpu(ncpus)
                        #print("ncpus is", ncpus)
                        print(ncpus, end='\t', file=f)
                        for size in range_multi(mintime, maxtime, factortime):
                                #print("time size is", size)
                                start=time.time()
                                for i in range(ntasks*ncpus):
                                        fut=starpu.task_submit(ret_fut=True)("func_test", size)
                                starpupy.task_wait_for_all()
                                end=time.time()
                                timing = end-start
                                print(size, "\t", timing/ncpus, end='\t', file=f)
                        print(end='\n', file=f)
        asyncio.run(main())

else:
        # return value is neither future nor handle
        for ncpus in range(mincpus, maxcpus+1, cpustep):
                starpupy.set_ncpu(ncpus)
                #print("ncpus is", ncpus)
                print(ncpus, end='\t', file=f)
                for size in range_multi(mintime, maxtime, factortime):
                        #print("time size is", size)
                        start=time.time()
                        for i in range(ntasks*ncpus):
                                fut=starpu.task_submit(ret_fut=False)("func_test", size)
                        starpupy.task_wait_for_all()
                        end=time.time()
                        timing = end-start
                        print(size, "\t", timing/ncpus, end='\t', file=f)
                print(end='\n', file=f)

#pr.disable()

f.close()
#pr.print_stats()
starpupy.shutdown()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# StarPU --- Runtime system for heterogeneous multicore architectures.
#
# Copyright (C) 2020-2022  Université de Bordeaux, CNRS (LaBRI UMR 5800), Inria
#
# StarPU is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or (at
# your option) any later version.
#
# StarPU is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See the GNU Lesser General Public License in COPYING.LGPL for more details.
#

import sys

"""
Convert event keys into event names
Running starpu_fxt_tool with the option -number-events produces a file number_events.data
This file contains the number of events for each event type.
Events are represented with their key.
To convert event keys to event names, call starpu_fxt_number_events_to_names.py
"""

# STARPU_FXT_EVENT_DEFINES is generated by configure and is the output of
# the following command:
# grep -E "#define\s+_STARPU_(MPI_)?FUT_" src/common/fxt.h mpi/src/starpu_mpi_fxt.h | grep 0x | grep -v 0x1 |cut -d : -f 2

fxt_codes_raw = """
#define _STARPU_FUT_WORKER_INIT_START	0x5100
#define _STARPU_FUT_WORKER_INIT_END	0x5101
#define	_STARPU_FUT_START_CODELET_BODY	0x5102
#define	_STARPU_FUT_END_CODELET_BODY	0x5103
#define _STARPU_FUT_JOB_PUSH		0x5104
#define _STARPU_FUT_JOB_POP		0x5105
#define _STARPU_FUT_UPDATE_TASK_CNT	0x5106
#define _STARPU_FUT_START_FETCH_INPUT_ON_TID	0x5107
#define _STARPU_FUT_END_FETCH_INPUT_ON_TID	0x5108
#define _STARPU_FUT_START_PUSH_OUTPUT_ON_TID	0x5109
#define _STARPU_FUT_END_PUSH_OUTPUT_ON_TID	0x5110
#define _STARPU_FUT_TAG		0x5111
#define _STARPU_FUT_TAG_DEPS	0x5112
#define _STARPU_FUT_TASK_DEPS		0x5113
#define _STARPU_FUT_DATA_COPY		0x5114
#define _STARPU_FUT_WORK_STEALING	0x5115
#define _STARPU_FUT_WORKER_DEINIT_START	0x5116
#define _STARPU_FUT_WORKER_DEINIT_END	0x5117
#define _STARPU_FUT_WORKER_SLEEP_START	0x5118
#define _STARPU_FUT_WORKER_SLEEP_END	0x5119
#define _STARPU_FUT_TASK_SUBMIT		0x511a
#define _STARPU_FUT_CODELET_DATA_HANDLE	0x511b
#define _STARPU_FUT_MODEL_NAME		0x511c
#define _STARPU_FUT_DATA_NAME		0x511d
#define _STARPU_FUT_DATA_COORDINATES	0x511e
#define _STARPU_FUT_HANDLE_DATA_UNREGISTER	0x511f
#define _STARPU_FUT_CODELET_DATA_HANDLE_NUMA_ACCESS	0x5120
#define	_STARPU_FUT_NEW_MEM_NODE	0x5122
#define	_STARPU_FUT_START_CALLBACK	0x5123
#define	_STARPU_FUT_END_CALLBACK	0x5124
#define	_STARPU_FUT_TASK_DONE		0x5125
#define	_STARPU_FUT_TAG_DONE		0x5126
#define	_STARPU_FUT_START_ALLOC		0x5127
#define	_STARPU_FUT_END_ALLOC		0x5128
#define	_STARPU_FUT_START_ALLOC_REUSE	0x5129
#define	_STARPU_FUT_END_ALLOC_REUSE	0x5130
#define	_STARPU_FUT_USED_MEM	0x512a
#define _STARPU_FUT_TASK_NAME	0x512b
#define _STARPU_FUT_DATA_WONT_USE	0x512c
#define _STARPU_FUT_TASK_COLOR	0x512d
#define _STARPU_FUT_DATA_DOING_WONT_USE	0x512e
#define _STARPU_FUT_TASK_LINE	0x512f
#define	_STARPU_FUT_START_MEMRECLAIM	0x5131
#define	_STARPU_FUT_END_MEMRECLAIM	0x5132
#define	_STARPU_FUT_START_DRIVER_COPY	0x5133
#define	_STARPU_FUT_END_DRIVER_COPY	0x5134
#define	_STARPU_FUT_START_DRIVER_COPY_ASYNC	0x5135
#define	_STARPU_FUT_END_DRIVER_COPY_ASYNC	0x5136
#define	_STARPU_FUT_START_PROGRESS_ON_TID	0x5137
#define	_STARPU_FUT_END_PROGRESS_ON_TID		0x5138
#define _STARPU_FUT_USER_EVENT		0x5139
#define _STARPU_FUT_SET_PROFILING	0x513a
#define _STARPU_FUT_TASK_WAIT_FOR_ALL	0x513b
#define _STARPU_FUT_EVENT		0x513c
#define _STARPU_FUT_THREAD_EVENT	0x513d
#define	_STARPU_FUT_CODELET_DETAILS	0x513e
#define	_STARPU_FUT_CODELET_DATA	0x513f
#define _STARPU_FUT_LOCKING_MUTEX	0x5140
#define _STARPU_FUT_MUTEX_LOCKED	0x5141
#define _STARPU_FUT_UNLOCKING_MUTEX	0x5142
#define _STARPU_FUT_MUTEX_UNLOCKED	0x5143
#define _STARPU_FUT_TRYLOCK_MUTEX	0x5144
#define _STARPU_FUT_RDLOCKING_RWLOCK	0x5145
#define _STARPU_FUT_RWLOCK_RDLOCKED	0x5146
#define _STARPU_FUT_WRLOCKING_RWLOCK	0x5147
#define _STARPU_FUT_RWLOCK_WRLOCKED	0x5148
#define _STARPU_FUT_UNLOCKING_RWLOCK	0x5149
#define _STARPU_FUT_RWLOCK_UNLOCKED	0x514a
#define _STARPU_FUT_LOCKING_SPINLOCK	0x514b
#define _STARPU_FUT_SPINLOCK_LOCKED	0x514c
#define _STARPU_FUT_UNLOCKING_SPINLOCK	0x514d
#define _STARPU_FUT_SPINLOCK_UNLOCKED	0x514e
#define _STARPU_FUT_TRYLOCK_SPINLOCK	0x514f
#define _STARPU_FUT_COND_WAIT_BEGIN	0x5150
#define _STARPU_FUT_COND_WAIT_END	0x5151
#define _STARPU_FUT_MEMORY_FULL		0x5152
#define _STARPU_FUT_DATA_LOAD 		0x5153
#define _STARPU_FUT_START_UNPARTITION_ON_TID 0x5154
#define _STARPU_FUT_END_UNPARTITION_ON_TID 0x5155
#define	_STARPU_FUT_START_FREE		0x5156
#define	_STARPU_FUT_END_FREE		0x5157
#define	_STARPU_FUT_START_WRITEBACK	0x5158
#define	_STARPU_FUT_END_WRITEBACK	0x5159
#define _STARPU_FUT_SCHED_COMPONENT_PUSH_PRIO 	0x515a
#define _STARPU_FUT_SCHED_COMPONENT_POP_PRIO 	0x515b
#define	_STARPU_FUT_START_WRITEBACK_ASYNC	0x515c
#define	_STARPU_FUT_END_WRITEBACK_ASYNC		0x515d
#define	_STARPU_FUT_HYPERVISOR_BEGIN    0x5160
#define	_STARPU_FUT_HYPERVISOR_END	0x5161
#define _STARPU_FUT_BARRIER_WAIT_BEGIN		0x5162
#define _STARPU_FUT_BARRIER_WAIT_END		0x5163
#define _STARPU_FUT_WORKER_SCHEDULING_START	0x5164
#define _STARPU_FUT_WORKER_SCHEDULING_END	0x5165
#define _STARPU_FUT_WORKER_SCHEDULING_PUSH	0x5166
#define _STARPU_FUT_WORKER_SCHEDULING_POP	0x5167
#define	_STARPU_FUT_START_EXECUTING	0x5168
#define	_STARPU_FUT_END_EXECUTING	0x5169
#define _STARPU_FUT_SCHED_COMPONENT_NEW		0x516a
#define _STARPU_FUT_SCHED_COMPONENT_CONNECT	0x516b
#define _STARPU_FUT_SCHED_COMPONENT_PUSH	0x516c
#define _STARPU_FUT_SCHED_COMPONENT_PULL	0x516d
#define _STARPU_FUT_TASK_SUBMIT_START	0x516e
#define _STARPU_FUT_TASK_SUBMIT_END	0x516f
#define _STARPU_FUT_TASK_BUILD_START	0x5170
#define _STARPU_FUT_TASK_BUILD_END	0x5171
#define _STARPU_FUT_TASK_MPI_DECODE_START	0x5172
#define _STARPU_FUT_TASK_MPI_DECODE_END		0x5173
#define _STARPU_FUT_TASK_MPI_PRE_START	0x5174
#define _STARPU_FUT_TASK_MPI_PRE_END	0x5175
#define _STARPU_FUT_TASK_MPI_POST_START	0x5176
#define _STARPU_FUT_TASK_MPI_POST_END	0x5177
#define _STARPU_FUT_TASK_WAIT_START	0x5178
#define _STARPU_FUT_TASK_WAIT_END	0x5179
#define _STARPU_FUT_TASK_WAIT_FOR_ALL_START	0x517a
#define _STARPU_FUT_TASK_WAIT_FOR_ALL_END	0x517b
#define _STARPU_FUT_HANDLE_DATA_REGISTER	0x517c
#define _STARPU_FUT_START_FETCH_INPUT	0x517e
#define _STARPU_FUT_END_FETCH_INPUT	0x517f
#define _STARPU_FUT_TASK_THROTTLE_START	0x5180
#define _STARPU_FUT_TASK_THROTTLE_END	0x5181
#define _STARPU_FUT_DATA_STATE_INVALID 0x5182
#define _STARPU_FUT_DATA_STATE_OWNER      0x5183
#define _STARPU_FUT_DATA_STATE_SHARED     0x5184
#define _STARPU_FUT_DATA_REQUEST_CREATED   0x5185
#define _STARPU_FUT_PAPI_TASK_EVENT_VALUE   0x5186
#define _STARPU_FUT_TASK_EXCLUDE_FROM_DAG   0x5187
#define _STARPU_FUT_TASK_END_DEP	0x5188
#define _STARPU_FUT_TASK_BUBBLE		0x5189
#define	_STARPU_FUT_START_PARALLEL_SYNC	0x518a
#define	_STARPU_FUT_END_PARALLEL_SYNC	0x518b
#define _STARPU_MPI_FUT_START				0x5201
#define _STARPU_MPI_FUT_STOP				0x5202
#define _STARPU_MPI_FUT_BARRIER				0x5203
#define _STARPU_MPI_FUT_ISEND_SUBMIT_BEGIN		0x5204
#define _STARPU_MPI_FUT_ISEND_SUBMIT_END		0x5205
#define _STARPU_MPI_FUT_IRECV_SUBMIT_BEGIN		0x5206
#define _STARPU_MPI_FUT_IRECV_SUBMIT_END		0x5207
#define _STARPU_MPI_FUT_ISEND_COMPLETE_BEGIN		0x5208
#define _STARPU_MPI_FUT_ISEND_COMPLETE_END		0x5209
#define _STARPU_MPI_FUT_DATA_SET_RANK			0x521a
#define _STARPU_MPI_FUT_IRECV_TERMINATED		0x521b
#define _STARPU_MPI_FUT_ISEND_TERMINATED		0x521c
#define _STARPU_MPI_FUT_TESTING_DETACHED_BEGIN		0x521d
#define _STARPU_MPI_FUT_TESTING_DETACHED_END		0x521e
#define _STARPU_MPI_FUT_TEST_BEGIN			0x521f
#define _STARPU_MPI_FUT_TEST_END			0x5220
#define _STARPU_MPI_FUT_IRECV_COMPLETE_BEGIN		0x520a
#define _STARPU_MPI_FUT_IRECV_COMPLETE_END		0x520b
#define _STARPU_MPI_FUT_SLEEP_BEGIN			0x520c
#define _STARPU_MPI_FUT_SLEEP_END			0x520d
#define _STARPU_MPI_FUT_DTESTING_BEGIN			0x520e
#define _STARPU_MPI_FUT_DTESTING_END			0x520f
#define _STARPU_MPI_FUT_UTESTING_BEGIN			0x5210
#define _STARPU_MPI_FUT_UTESTING_END			0x5211
#define _STARPU_MPI_FUT_UWAIT_BEGIN			0x5212
#define _STARPU_MPI_FUT_UWAIT_END			0x5213
#define _STARPU_MPI_FUT_POLLING_BEGIN			0x5214
#define _STARPU_MPI_FUT_POLLING_END			0x5215
#define _STARPU_MPI_FUT_DRIVER_RUN_BEGIN		0x5216
#define _STARPU_MPI_FUT_DRIVER_RUN_END			0x5217
#define _STARPU_MPI_FUT_DATA_SET_TAG			0x5218
#define _STARPU_MPI_FUT_IRECV_NUMA_NODE			0x5219
#define _STARPU_MPI_FUT_ISEND_NUMA_NODE			0x5221
#define _STARPU_MPI_FUT_CHECKPOINT_BEGIN            	0x5222
#define _STARPU_MPI_FUT_CHECKPOINT_END              	0x5223
"""

PROGNAME=sys.argv[0]

number_events_path = None

def usage():
    print("Convert event keys in number_events.data to event names")
    print("")
    print("Usage: %s <number_events.data path>" % PROGNAME)
    print("")
    print("Options:")
    print("	-h, --help          display this help and exit")
    print("	-v, --version       output version information and exit")
    print("")
    print("Report bugs to <starpu-devel@inria.fr>")
    sys.exit(1)

if len(sys.argv) == 2:
    if sys.argv[1] == '-v' or sys.argv[1] == '--version':
        print("%s (StarPU) 1.4.5" % PROGNAME)
        sys.exit(0)
    elif sys.argv[1] == '-h' or sys.argv[1] == '--help':
        usage()
    else:
        number_events_path = sys.argv[1]
else:
    usage()

# Process fxt_code_raw content to ease the conversion:
fxt_codes = dict()
for line in fxt_codes_raw.split("\n"):
    elements = line.split()

    if len(elements) == 3:
        key = int(elements[2][2:], 16)
        assert key not in fxt_codes

        fxt_codes[key] = elements[1]


# Convert content of the file:
nb_events = 0

with open(number_events_path, 'r') as f:
    for line in f:
        elements = line.split()
        if len(elements) == 2:
            key = int(elements[0][2:], 16)
            nb = int(elements[1])
            nb_events += nb
            if key in fxt_codes:
                print("%12d    %s" % (nb, fxt_codes[key]))
            else:
                print("%12d    %s" % (nb, elements[0]))

print("       TOTAL:   %d" % nb_events)



# StarPU --- Runtime system for heterogeneous multicore architectures.
#
# Copyright (C) 2020-2023  Universit'e de Bordeaux, CNRS (LaBRI UMR 5800), Inria
#
# StarPU is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or (at
# your option) any later version.
#
# StarPU is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See the GNU Lesser General Public License in COPYING.LGPL for more details.
#

try:
    import numpy as np
except (ModuleNotFoundError, ImportError):
    print("Can't find \"Python3 NumPy\" module (consider running \"pip3 install numpy\" or refer to https://numpy.org/install/)")
    np = None

import starpu
from starpu import starpupy
from starpu import Handle
from starpu import HandleNumpy
import asyncio
import time
import array
import struct

try:
        starpu.init()
except Exception as e:
        print(e)
        exit(77)

def await_fut(fut):
    return fut.result()

if starpupy.worker_get_count_by_type(starpu.STARPU_MPI_MS_WORKER) >= 1 or starpupy.worker_get_count_by_type(starpu.STARPU_TCPIP_MS_WORKER) >= 1:
	print("This program does not work in MS mode")
	starpu.shutdown()
	exit(77)

def show(x, y):
    print("Function printing:", x, y)

def add(x, y):
	print ("Example add(x, y):")
	return x + y

# create Handle objects
x=2
y=3
x_h = Handle(x)
y_h = Handle(y)

print("*************************")
print("constant handle:")
print("*************************")
# show function returns Handle
ret_h1 = starpu.task_submit(ret_handle=True)(show, "first argument is:", x_h)
print("show function returns:", ret_h1.get())

# return value is Handle
res1 = starpu.task_submit(ret_handle=True)(add, x_h, y_h)
print("result of Handle(2)+Handle(3) is:", res1.get())

# return value is Handle
res2 = starpu.task_submit(ret_handle=True)(add, res1, y_h)
print("result of res1+Handle(3) is:", res2.get())

# show function returns Handle
ret_h2 = starpu.task_submit(ret_handle=True)(show, res1, res2)

print("*************************************")
print("constant handle return in parameter:")
print("*************************************")

ret = Handle(0)
print("before calling function, ret value is:", ret.get())
# return value as parameter
ret_n = starpu.task_submit(ret_param=True)(add, ret, x_h, y_h)
print("result of Handle(2)+Handle(3) is:", ret.get())
print("return value of task_submit is:", ret_n)
assert ret.get() == x+y

x_h.unregister()
y_h.unregister()

ret_h1.unregister()
ret_h2.unregister()
res2.unregister()
ret.unregister()

if np is not None:
    ##############################################################################################
    print("*************************")
    print("Numpy array handle:")
    print("*************************")
    def scal(x, t):
        for i in range(len(t)):
            t[i] = t[i] * x
        print ("Example scal(scalar, array):")

    t = np.arange(10)

    # create Handle object for Numpy array
    t_h = Handle(t)

    # return value is Handle
    res3 = starpu.task_submit(ret_handle=True)(scal, 2, t_h)
    print("result of scal(2, Handle(np.arange(10)) is:", t_h.get())

    # show function returns Future
    async def main():
        res_fut1 = starpu.task_submit()(show, res1, t_h)
        await(res_fut1)
    asyncio.run(main())

    t_h.unregister()
    res1.unregister()
    res3.unregister()

    ######################
    def arr_add(a,b):
        for i in range(np.size(a)):
            a[i] = a[i] + b[i]

    a = np.array([1, 2, 3])
    b = np.array([4, 5, 6])

    # create Handle objects
    a_h = Handle(a)
    b_h = Handle(b)

    # two array element addition
    res4 = starpu.task_submit(ret_handle=True)(arr_add, a_h, b_h)
    print("result of adding two Handle(numpy.array) is:", a_h.get())

    a_h.unregister()
    b_h.unregister()

    res4.unregister()

    #######################
    def multi(x, y):
        print ("Example multi(x, y):")
        np.multiply(x, y, out=x)

    c = np.array([[1, 2], [3, 4]])
    d = np.array([[2, 2], [2, 2]])

    # create Handle objects
    c_h = Handle(c)
    d_h = Handle(d)

    # two array element multiplication
    res5 = starpu.task_submit(ret_handle=True)(multi, c_h, d_h)
    print("result of multiplying two Handle(numpy.array) is:", c_h.get())

    ########################
    @starpu.access(x="RW")
    def matrix_multi(x, y):
        print ("Example matrix_multi(x, y):")
        np.dot(x, y, out=x)

    # two array matrix multiplication
    res6 = starpu.task_submit(ret_handle=True)(matrix_multi, c_h, d_h)
    print("result of matrix multiplying two Handle(numpy.array) is:", c_h.get())

    # two array matrix multiplication (inverse order)
    res7 = starpu.task_submit(ret_handle=True)(matrix_multi, d_h, c_h)
    print("result of matrix multiplying two Handle(numpy.array) is:", d_h.get())

    c_h.unregister()
    d_h.unregister()

    res5.unregister()
    res6.unregister()
    res7.unregister()

    ###################################empty Numpy array handle#####################################
    print("*************************")
    print("empty Numpy array handle:")
    print("*************************")
    a1 = np.array([1, 2, 3, 4])
    a2 = np.array([[1, 2, 3], [4, 5, 6]])
    a3 = np.array([[[1, 2, 3], [4, 5, 6]],[[7, 8, 9], [10, 11, 12]]])

    # create Handle objects
    a1_h = Handle(a1)
    a2_h = Handle(a2)
    a3_h = Handle(a3)

    a1_r = a1_h.acquire(mode='R')
    print("original 1-dimension array is:", a1_r)
    a1_h.release()
    a2_r = a2_h.acquire(mode='R')
    print("original 2-dimension array is:", a2_r)
    a2_h.release()
    a3_r = a3_h.acquire(mode='R')
    print("original 3-dimension array is:", a3_r)
    a3_h.release()

    @starpu.access(b="W")
    def assign(a,b):
        for i in range(min(np.size(a), np.size(b))):
            b[i] = a[i]

    @starpu.access(b="W")
    def assign2(a,b):
        for i in range(min(np.size(a,0), np.size(b,0))):
            for j in range(min(np.size(a,1), np.size(b,1))):
                b[i][j] = a[i][j]

    @starpu.access(b="W")
    def assign3(a,b):
        for i in range(min(np.size(a,0), np.size(b,0))):
            for j in range(min(np.size(a,1), np.size(b,1))):
                for k in range(min(np.size(a,2), np.size(b,2))):
                    b[i][j][k] = a[i][j][k]

    # generate empty arrays Handle object using HandleNumpy
    # 1-dimension
    e1_h = HandleNumpy(a1.shape, a1.dtype)

    res8 = starpu.task_submit(ret_handle=True)(assign, a1_h, e1_h)
    e1_r = e1_h.acquire(mode='RW')
    print("assigned 1-dimension array is:", e1_r)
    # e1_h is writeable, we modify the first element
    e1_r[0] = 100
    print("the first element of 1-dimension array is modified to 100:", e1_r)
    e1_h.release()

    # 2-dimension
    e2_h = HandleNumpy(a2.shape, a2.dtype)
    res9 = starpu.task_submit(ret_handle=True)(assign2, a2_h, e2_h)
    e2_r = e2_h.acquire(mode='R')
    print("assigned 2-dimension array is", e2_r)
    e2_h.release()

    # 3-dimension
    e3_h = HandleNumpy(a3.shape, a3.dtype)
    res10 = starpu.task_submit(ret_handle=True)(assign3, a3_h, e3_h)
    e3_r = e3_h.acquire(mode='R')
    print("assigned 3-dimension array is", e3_r)
    e3_h.release()

    a1_h.unregister()
    a2_h.unregister()
    a3_h.unregister()
    e1_h.unregister()
    e2_h.unregister()
    e3_h.unregister()

    res8.unregister()
    res9.unregister()
    res10.unregister()

##################################bytes handle############################################
print("*************************")
print("bytes handle:")
print("*************************")
bt1 = bytes([1,2])
bt2 = bytes([3,4])

bt1_h = Handle(bt1)
bt2_h = Handle(bt2)

bt1_r = bt1_h.acquire(mode='R')
print("first bytes object is", bt1_r)
bt1_h.release()

bt2_r = bt2_h.acquire(mode='R')
print("second bytes object is", bt2_r)
bt2_h.release()

ret_bt1 = starpu.task_submit(ret_handle=True)(add, bt1_h, bt2_h)
print("result of appending two bytes: ", ret_bt1.get())

def bytes_add(x, y):
	z = bytearray(len(x))
	for i in range (len(x)):
		z[i] = x[i] + y[i]
	return bytes(z)

ret_bt2 = starpu.task_submit(ret_handle=True)(bytes_add, bt1_h, bt2_h)
print("result of adding two bytes elements: ", ret_bt2.get())

bt1_h.unregister()
bt2_h.unregister()

ret_bt1.unregister()
ret_bt2.unregister()

####################################bytearray handle#########################################
print("*************************")
print("bytearray handle:")
print("*************************")
bta1 = bytearray([1,2])
bta2 = bytearray([3,4])

bta1_h = Handle(bta1)
bta2_h = Handle(bta2)

bta1_r = bta1_h.acquire(mode='RW')
print("first bytearray object is", bta1_r)
bta1[0] = 0
bta1_h.release()
bta11_r = bta1_h.acquire(mode='R')
print("first bytearray object is modified", bta11_r)
bta1_h.release()

bta2_r = bta2_h.acquire(mode='R')
print("second bytearray object is", bta2_r)
bta2_h.release()

def bytearray_add(x, y):
	z = bytearray(len(x))
	for i in range (len(x)):
		z[i] = x[i] + y[i]
	return z

ret_bta1 = starpu.task_submit(ret_handle=True)(bytearray_add, bta1_h, bta2_h)
print("result of adding two bytearray elements: ", ret_bta1.get())

bta1_h.unregister()
bta2_h.unregister()

ret_bta1.unregister()

##################################array.array handle##########################################
print("*************************")
print("array.array handle:")
print("*************************")
arr1 = array.array('i', [1, 2, 3, 4])
arr2 = array.array('i', [2, 2, 2, 2])
arr3 = array.array('f', [4.5, 5.5, 6.5])

arr4 = array.array('u', 'hello')

def arrarr_add(x, y):
	for i in range (len(x)):
		x[i] = x[i] + y[i]
	#time.sleep(1)
	return x

def arrarr_multi(x, y):
	for i in range (len(x)):
		x[i] = x[i] * y[i]
	return x

def arrarr_scal(x, s):
	for i in range (len(x)):
		x[i] = x[i] * s
	return x

arr1_h = Handle(arr1)
arr1_r = arr1_h.acquire(mode='RW')
print("first array.array object is", arr1_r)
arr1[0] = 0
arr1_h.release()
arr11_r = arr1_h.acquire(mode='R')
print("first array.array object is modified", arr11_r)
arr1_h.release()

arr2_h = Handle(arr2)
arr2_r = arr2_h.acquire(mode='R')
print("second array.array object is", arr2_r)
arr2_h.release()

arr3_h = Handle(arr3)
arr3_r = arr3_h.acquire(mode='R')
print("third array.array object is", arr3_r)
arr3_h.release()

arr4_h = Handle(arr4)
arr4_r = arr4_h.acquire(mode='R')
print("fourth array.array object is", arr4_r)
arr4_h.release()

ret_arr1 = starpu.task_submit(ret_handle=True)(arrarr_add, arr1_h, arr2_h)
print("result of adding two array.array elements: ", ret_arr1.get())

ret_arr2 = starpu.task_submit(ret_handle=True)(arrarr_multi, arr1_h, arr2_h)
print("result of multiplying two array.array elements: ", ret_arr2.get())

ret_arr3 = starpu.task_submit(ret_handle=True)(arrarr_scal, arr3_h, 2)
print("result of multiplying array.array element by a scalar: ", ret_arr3.get())

arr1_h.unregister()
arr2_h.unregister()
arr3_h.unregister()
arr4_h.unregister()

ret_arr1.unregister()
ret_arr2.unregister()
ret_arr3.unregister()

##################################memoryview handle###########################################
print("*************************")
print("memoryview handle:")
print("*************************")
m1 = memoryview(bytearray("hello", 'utf-8'))
m1_tb = m1.tobytes()
print("m1 to bytes is", m1_tb)

m2 = memoryview(array.array('i', [1, 2, 3, 4]))
m2_tl = m2.tolist()
print("m2 to list is", m2_tl)

m3 = memoryview(array.array('u', 'hello'))

m1_h = Handle(m1)
print("m1 is", m1_h.acquire(mode='RW'))
m1[0] = 100
m1_h.release()

print("m1 to bytes after modifying is", m1_tb)
print("m1 after modifying is", m1_h.acquire(mode='RW'))
m1_h.release()

m2_h = Handle(m2)
print("m2 is", m2_h.acquire(mode='R'))
m2_h.release()

m3_h = Handle(m3)
print("m3 is", m3_h.acquire(mode='R'))
m3_h.release()

# multi dimension
def mem_show(x):
	print("memory is", x)

buf = struct.pack("L"*12, *list(range(12)))
x = memoryview(buf)
# 2-dimension
y = x.cast('L', shape=[3,4])
# 3-dimension
z = x.cast('L', shape=[2,3,2])
print(y.tolist())
print(z.tolist())

y_h = Handle(y)
ret_m1 = starpu.task_submit(ret_handle=True)(mem_show, y_h)

print("y is", y_h.acquire(mode='R'))
y_h.release()

z_h = Handle(z)
ret_m2 = starpu.task_submit(ret_handle=True)(mem_show, z_h)

print("z is", z_h.acquire(mode='R'))
z_h.release()

m1_h.unregister()
m2_h.unregister()
m3_h.unregister()

y_h.unregister()
z_h.unregister()

ret_m1.unregister()
ret_m2.unregister()

if np is not None:
    #####################################access mode annotation###################################
    print("*************************")
    print("access mode annotation:")
    print("*************************")
    a = np.array([1, 2, 3, 4])
    a_h = Handle(a)
    e_h = HandleNumpy(a.shape, a.dtype)

    a_r = a_h.acquire(mode='R')
    print("original array is:", a_r)
    a_h.release()

    ######access#####
    print("------------------")
    print("access decorator:")
    print("------------------")
    @starpu.access(a="R", b="W")
    def assign(a,b):
        for i in range(min(np.size(a), np.size(b))):
            b[i]=a[i]

    res11 = starpu.task_submit(ret_handle=True)(assign, a_h, e_h)

    e_r = e_h.acquire(mode='RW')
    print("assigned 1-dimension array is:", e_r)
    e_h.release()

    ######delayed#######
    print("------------------")
    print("delayed decorator:")
    print("------------------")
    @starpu.delayed(ret_handle=True, a="R", b="W")
    def assign(a,b):
        for i in range(min(np.size(a), np.size(b))):
            b[i]=a[i]

    res12 = assign(a_h, e_h)

    e_r = e_h.acquire(mode='RW')
    print("assigned 1-dimension array is:", e_r)
    e_h.release()

    ######set access######
    print("------------------")
    print("access function:")
    print("------------------")
    def assign(a,b):
        for i in range(min(np.size(a), np.size(b))):
            b[i]=a[i]

    assign_access=starpu.set_access(assign, a="R", b="W")
    res13 = starpu.task_submit(ret_handle=True)(assign_access, a_h, e_h)

    e_r = e_h.acquire(mode='RW')
    print("assigned 1-dimension array is:", e_r)
    e_h.release()

    a_h.unregister()
    e_h.unregister()

    res11.unregister()
    res12.unregister()
    res13.unregister()

    #######################Numpy without explicit handle############################
    print("*******************************")
    print("Numpy without explicit handle:")
    print("*******************************")
    arrh1 = np.array([1, 2, 3])
    arrh2 = np.array([4, 5, 6])

    @starpu.access(a="RW", b="R")
    def np_add(a, b):
        #time.sleep(2)
        for i in range(np.size(a)):
            a[i] = a[i] + b[i]

    print("First argument before task submitting is", starpu.acquire(arrh1, mode='R'))
    #a[0]=100
    starpu.release(arrh1)
    # without explicit handle
    res14 = starpu.task_submit(ret_handle=True)(np_add, arrh1, arrh2)

    print("First argument after task submitting is", starpu.acquire(arrh1, mode='R'))
    starpu.release(arrh1)

    # it's mandatory to call unregister when the argument is no longer needed to access, but it's not obligatory, calling starpupy.shutdown() in the end is enough, which will unregister all no-explicit handle
    starpu.unregister(arrh1)

    res14.unregister()

    #######################Numpy without using handle###############################
    print("*******************************")
    print("Numpy without using handle:")
    print("*******************************")
    npa1 = np.array([1, 2, 3])
    npa2 = np.array([4, 5, 6])

    print("First argument before task submitting is", npa1)
    # without using handle, set option arg_handle to False
    res15 = starpu.task_submit(arg_handle=False, ret_handle=True)(np_add, npa1, npa2)
    print("First argument after task submitting is", npa1)
    #print("The addition result is", res15.get())

    res15.unregister()

#########################

starpu.shutdown()



# StarPU --- Runtime system for heterogeneous multicore architectures.
#
# Copyright (C) 2020-2023  Universit'e de Bordeaux, CNRS (LaBRI UMR 5800), Inria
#
# StarPU is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or (at
# your option) any later version.
#
# StarPU is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See the GNU Lesser General Public License in COPYING.LGPL for more details.
#

try:
    import numpy as np
except (ModuleNotFoundError, ImportError):
	print("Can't find \"Python3 NumPy\" module (consider running \"pip3 install numpy\" or refer to https://numpy.org/install/)")
	exit(77)

import starpu
from starpu import starpupy
import asyncio

try:
        starpu.init()
except Exception as e:
        print(e)
        exit(77)

def await_fut(fut):
    return fut.result()

###############################################################################

def scal(x, t):
	for i in range(len(t)):
		t[i] = t[i] * x
	print ("Example scal(scalar, array):")
	return t

def add(x, y):
	print ("Example add(array, array):")
	return x + y

def multi(x, y):
	print ("Example multi(array, array):")
	return x * y

def matrix_multi(x, y):
	print ("Example matrix_multi(array, array):")
	return x @ y

t = np.arange(10)

a = np.array([1, 2, 3])
b = np.array([4, 5, 6])

c = np.array([[1, 2], [3, 4]])
d = np.array([[2, 2], [2, 2]])

async def main():
    fut1 = starpu.task_submit()(scal, 2, t)
    res1 = await(fut1)
    print("The result is", res1)

    # two array element addition
    fut2 = starpu.task_submit()(add, a, b)
    res2 = await(fut2)
    print("The result is", res2)

    # two array element multiplication
    fut3 = starpu.task_submit()(multi, c, d)
    res3 = await(fut3)
    print("The result is", res3)

    # two array matrix multiplication
    fut4 = starpu.task_submit()(matrix_multi, c, d)
    res4 = await(fut4)
    print("The result is", res4)

    # two array matrix multiplication (inverse order)
    fut5 = starpu.task_submit()(matrix_multi, d, c)
    res5 = await(fut5)
    print("The result is", res5)


try:
        asyncio.run(main())
except starpupy.error as e:
        print("No worker to execute the job")
        starpupy.shutdown()
        exit(77)

starpu.shutdown()



# StarPU --- Runtime system for heterogeneous multicore architectures.
#
# Copyright (C) 2021  Université de Bordeaux, CNRS (LaBRI UMR 5800), Inria
#
# StarPU is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or (at
# your option) any later version.
#
# StarPU is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See the GNU Lesser General Public License in COPYING.LGPL for more details.
#
import json
import matplotlib.pyplot as plt
import test_handle_bench

file1 = open('handle_perf1.txt', 'r')
js1 = file1.read()
retfut_dict = json.loads(js1)
#print(retfut_dict)
program_submit1 = [x*1000 for x in retfut_dict['program_submit']]
program_await1 = [x*1000 for x in retfut_dict['program_await']]

file2 = open('handle_perf2.txt', 'r')
js2 = file2.read()
rethandle_dict = json.loads(js2)
#print(rethandle_dict)
program_submit2 = [x*1000 for x in rethandle_dict['program_submit']]
program_await2 = [x*1000 for x in rethandle_dict['program_await']]

file3 = open('handle_perf3.txt', 'r')
js3 = file3.read()
nostarpu_dict = json.loads(js3)
#print(nostarpu_dict)
program_submit3 = [x*1000 for x in nostarpu_dict['program_submit']]

file_std = open('handle_perf_std.txt', 'r')
js_std = file_std.read()
dict_std = json.loads(js_std)

file1.close()
file2.close()
file3.close()
file_std.close()

std11 = dict_std['list_std11']
std12 = dict_std['list_std12']
std21 = dict_std['list_std21']
std22 = dict_std['list_std22']
std3 = dict_std['list_std3']

plt.subplot(2, 1, 1)
plt.xticks(fontsize=15)
plt.yticks(fontsize=15)
plt.xscale("log")
plt.yscale("log")
plt.errorbar([i for i in test_handle_bench.list_size], program_submit1, yerr=std11, fmt='+-', ecolor='r', color='r', elinewidth=1, capsize=3, linewidth=1, label='using StarPU and returning future object')
plt.errorbar([i for i in test_handle_bench.list_size], program_submit2, yerr=std21, fmt='+-', ecolor='b', color='b', elinewidth=1, capsize=3, linewidth=1, label='using StarPU and returning handle object')
plt.errorbar([i for i in test_handle_bench.list_size], program_submit3, yerr=std3, fmt='+-',ecolor='y', color='y', elinewidth=1, capsize=3, linewidth=1, label='using numpy.add function')

plt.legend(loc='upper left', fontsize=15)
plt.xlabel("Numpy array size (# of elements)", fontsize=15)
plt.ylabel("Program execution time (ms)", fontsize=15)

plt.subplot(2, 1, 2)
plt.xticks(fontsize=15)
plt.yticks(fontsize=15)
plt.xscale("log")
plt.yscale("log")
plt.errorbar([i for i in test_handle_bench.list_size], program_await1, yerr=std12, fmt='+-',ecolor='r', color='r', elinewidth=1, capsize=3, linewidth=1, label='using StarPU and returning future object')
plt.errorbar([i for i in test_handle_bench.list_size], program_await2, yerr=std22, fmt='+-',ecolor='b', color='b', elinewidth=1, capsize=3, linewidth=1, label='using StarPU and returning handle object')

plt.legend(loc='upper left', fontsize=15)
plt.xlabel("Numpy array size (# of elements)", fontsize=15)
plt.ylabel("Program await time (ms)", fontsize=15)

plt.show()
#plt.savefig("starpupy_handle_perf.png")
#plt.savefig("starpupy_handle_perf.eps")




# StarPU --- Runtime system for heterogeneous multicore architectures.
#
# Copyright (C) 2020-2023  Universit'e de Bordeaux, CNRS (LaBRI UMR 5800), Inria
#
# StarPU is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or (at
# your option) any later version.
#
# StarPU is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See the GNU Lesser General Public License in COPYING.LGPL for more details.
#

from . import starpupy
from .delay import *
from .handle_access import *
#from . import joblib
from .intermedia import *

import asyncio
import concurrent.futures
try:
    import numpy as np
    has_numpy=True
except:
    has_numpy=False

async def wait_for_fut(fut):
	return await fut

async def asyncio_wait_for_fut(fut):
	return await fut

def concurrent_futures_wait_for_fut(fut):
	return fut.result()

#class handle
class Handle(object):

	def __init__(self, obj, retval=False):
		self.obj=obj
		self.obj_id=id(self.obj)
		self.retval=retval
		self.handle_cap=starpupy.starpupy_data_register(self.obj, self)

	def get_capsule(self):
		return self.handle_cap

	def get_obj_id(self):
		return self.obj_id

	def get_retval(self):
		return self.retval

	# get PyObject
	def get(self):
		return starpupy.starpupy_get_object(self.handle_cap)

	# get array object
	def acquire(self, mode='R'):
		return starpupy.starpupy_acquire_handle(self.handle_cap, mode)

	# release
	def release(self):
		return starpupy.starpupy_release_handle(self.handle_cap)

	# unregister
	def unregister(self):
		return starpupy.starpupy_data_unregister(self)

	# unregister_submit
	def unregister_submit(self):
		return starpupy.starpupy_data_unregister_submit(self)

	# partition
	def partition(self, nchildren, dim, chunks_list=[]):
		return starpupy.starpupy_data_partition(self.handle_cap, nchildren, dim, chunks_list)

	# get partition size
	def get_partition_size(self, handle_list):
		return starpupy.starpupy_get_partition_size(self.handle_cap, handle_list)

	# unpartition
	def unpartition(self, handle_list, nchildren):
		return starpupy.starpupy_data_unpartition(self.handle_cap, handle_list, nchildren)

def new_empty_numpy(shape, dtype):
	return np.empty(shape, dtype)

#class handle
class HandleNumpy(Handle):
	if has_numpy:
		def __init__(self, shape, dtype=np.dtype('float64'), retval=False):
			self.dtype=dtype
			self.retval=retval
			self.obj=new_empty_numpy(shape, self.dtype)
			self.obj_id=id(self.obj)
			self.handle_cap=starpupy.starpupy_data_register(self.obj, self)


#detect class handle
class Handle_token(object):
	pass

#this dict contains all handle objects of mutable Python objects
handle_dict={}
def handle_dict_set_item(obj, handle):
	assert handle_dict.get(id(obj))==None
	handle_dict[id(obj)]=handle
	return handle_dict

#this set contains all handle objects of immutable Python objects
handle_set=set()



# StarPU --- Runtime system for heterogeneous multicore architectures.
#
# Copyright (C) 2020-2024  Universit'e de Bordeaux, CNRS (LaBRI UMR 5800), Inria
#
# StarPU is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or (at
# your option) any later version.
#
# StarPU is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See the GNU Lesser General Public License in COPYING.LGPL for more details.
#

try:
    import numpy as np
except (ModuleNotFoundError, ImportError):
    print("Can't find \"Python3 NumPy\" module (consider running \"pip3 install numpy\" or refer to https://numpy.org/install/)")
    np = None

import starpu
try:
    import starpu.joblib
except (ModuleNotFoundError, ImportError):
    print("Can't find starpu.joblib\" module (consider running \"pip3 install joblib\")")
    exit(77)
from starpu import starpupy
import time
import asyncio
from math import sqrt
from math import log10
import sys

try:
        starpu.init()
except Exception as e:
        print(e)
        exit(77)

def await_fut(fut):
    return fut.result()

#generate a list to store functions
g_func=[]

#function no input no output print hello world
def hello():
	print ("Example 1: Hello, world!")
g_func.append(starpu.joblib.delayed(hello)())

#function no input no output
def func1():
	print ("Example 2: This is a function no input no output")
g_func.append(starpu.joblib.delayed(func1)())

#function no input return a value
def func2():
	print ("Example 3:")
	return 12
g_func.append(starpu.joblib.delayed(func2)())

#function has 2 int inputs and 1 int output
def exp(a,b):
	res_exp=a**b
	print("Example 4: The result of ",a,"^",b,"is",res_exp)
	return res_exp
g_func.append(starpu.joblib.delayed(exp)(2, 3))

#function has 4 float inputs and 1 float output
def add(a,b,c,d):
	res_add=a+b+c+d
	print("Example 5: The result of ",a,"+",b,"+",c,"+",d,"is",res_add)
	return res_add
g_func.append(starpu.joblib.delayed(add)(1.2, 2.5, 3.6, 4.9))

#function has 2 int inputs 1 float input and 1 float output 1 int output
def sub(a,b,c):
	res_sub1=a-b-c
	res_sub2=a-b
	print ("Example 6: The result of ",a,"-",b,"-",c,"is",res_sub1,"and the result of",a,"-",b,"is",res_sub2)
	return res_sub1, res_sub2
g_func.append(starpu.joblib.delayed(sub)(6, 2, 5.9))

##########functions of array calculation###############
def scal(a, t):
	for i in range(len(t)):
		t[i]=t[i]*a
	return t

@starpu.access(t="RW")
def scal_np(a, t):
	for i in range(len(t)):
		t[i]=t[i]*a

@starpu.access(t1="RW")
def add_scal(a, t1, t2):
	for i in range(len(t1)):
		t1[i]=t1[i]*a+t2[i]
	#return t1

@starpu.access(t="RW")
def scal_arr(a, t):
	for i in range(len(t)):
		t[i]=t[i]*a[i]

def multi(a,b):
	res_multi=a*b
	return res_multi

def multi_2arr(a, b):
    for i in range(len(a)):
    	a[i]=a[i]*b[i]
    return a

@starpu.access(a="RW")
def multi_2np(a, b):
    for i in range(len(a)):
    	a[i]=a[i]*b[i]

def multi_list(l):
	res = []
	for (a,b) in l:
		res.append(a*b)
	return res

@starpu.access(t="RW")
def log10_arr(t):
	for i in range(len(t)):
		t[i]=log10(t[i])

########################################################

displayPlot=False
listX=[10, 100]
for arg in sys.argv[1:]:
        if arg == "-long":
                listX = [10, 100, 1000, 10000, 100000, 1000000, 10000000]
        if arg == "-plot":
                displayPlot=True

if np is not None:
    #################scikit test###################
    # DEFAULT_JOBLIB_BACKEND = starpu.joblib.get_active_backend()[0].__class__
    # class MyBackend(DEFAULT_JOBLIB_BACKEND):  # type: ignore
    #         def __init__(self, *args, **kwargs):
    #                 self.count = 0
    #                 super().__init__(*args, **kwargs)

    #         def start_call(self):
    #                 self.count += 1
    #                 return super().start_call()

    # starpu.joblib.register_parallel_backend('testing', MyBackend)

    # with starpu.joblib.parallel_backend("testing") as (ba, n_jobs):
    # 	print("backend and n_jobs is", ba, n_jobs)
    ###############################################

    N=100
    # A=np.arange(N)
    # B=np.arange(N)
    # a=np.arange(N)
    # b=np.arange(N, 2*N, 1)

    for x in listX:
        for X in range(x, x*10, x):
            #print("X=",X)
            try :
                starpu.joblib.Parallel(mode="normal", n_jobs=-1, perfmodel="log_list")(starpu.joblib.delayed(log10)(i+1)for i in range(X))
                A=np.arange(1,X+1,1)
                starpu.joblib.Parallel(mode="normal", n_jobs=-1, perfmodel="log_arr")(starpu.joblib.delayed(log10_arr)(A))
            except starpupy.error as e:
                print("No worker to execute the job")
                exit(77)

    print("************************")
    print("parallel Normal version:")
    print("************************")
    print("--(sqrt)(i**2)for i in range(N)")
    start_exec1=time.time()
    start_cpu1=time.process_time()
    starpu.joblib.Parallel(mode="normal", n_jobs=-1, perfmodel="sqrt")(starpu.joblib.delayed(sqrt)(i**2)for i in range(N))
    end_exec1=time.time()
    end_cpu1=time.process_time()
    print("the program execution time is", end_exec1-start_exec1)
    print("the cpu execution time is", end_cpu1-start_cpu1)

    print("--(multi)(i,j) for i,j in zip(a,b)")
    a=np.arange(N)
    b=np.arange(N, 2*N, 1)
    start_exec2=time.time()
    start_cpu2=time.process_time()
    starpu.joblib.Parallel(mode="normal", n_jobs=-1, perfmodel="multi")(starpu.joblib.delayed(multi)(i,j) for i,j in zip(a,b))
    end_exec2=time.time()
    end_cpu2=time.process_time()
    print("the program execution time is", end_exec2-start_exec2)
    print("the cpu execution time is", end_cpu2-start_cpu2)

    print("--(scal_arr)((i for i in b), A)")
    A=np.arange(N)
    b=np.arange(N, 2*N, 1)
    print("The input array is", A)
    start_exec3=time.time()
    start_cpu3=time.process_time()
    starpu.joblib.Parallel(mode="normal", n_jobs=-1, perfmodel="scal_arr")(starpu.joblib.delayed(scal_arr)((i for i in b), A))
    end_exec3=time.time()
    end_cpu3=time.process_time()
    print("The return array is", A)
    print("the program execution time is", end_exec3-start_exec3)
    print("the cpu execution time is", end_cpu3-start_cpu3)

    print("--(multi_list)((i,j) for i,j in zip(a,b))")
    a=np.arange(N)
    b=np.arange(N, 2*N, 1)
    start_exec4=time.time()
    start_cpu4=time.process_time()
    starpu.joblib.Parallel(mode="normal", n_jobs=-1, perfmodel="multi_list")(starpu.joblib.delayed(multi_list)((i,j) for i,j in zip(a,b)))
    end_exec4=time.time()
    end_cpu4=time.process_time()
    print("the program execution time is", end_exec4-start_exec4)
    print("the cpu execution time is", end_cpu4-start_cpu4)

    print("--(multi_2arr)((i for i in a), (j for j in b))")
    a=np.arange(N)
    b=np.arange(N, 2*N, 1)
    start_exec5=time.time()
    start_cpu5=time.process_time()
    starpu.joblib.Parallel(mode="normal", n_jobs=-1, perfmodel="multi_2arr")(starpu.joblib.delayed(multi_2arr)((i for i in a), (j for j in b)))
    end_exec5=time.time()
    end_cpu5=time.process_time()
    print("the program execution time is", end_exec5-start_exec5)
    print("the cpu execution time is", end_cpu5-start_cpu5)

    print("--(multi_2np)(A, B)")
    # A=np.arange(N)
    # B=np.arange(N, 2*N, 1)
    n, m = 4, 5
    A = np.arange(n*m).reshape(n, m)
    B = np.arange(n*m, 2*n*m, 1).reshape(n, m)
    print("The input arrays are A", A, "B", B)
    start_exec6=time.time()
    start_cpu6=time.process_time()
    starpu.joblib.Parallel(mode="normal", n_jobs=-1, perfmodel="multi_2arr")(starpu.joblib.delayed(multi_2np)(A, B))
    end_exec6=time.time()
    end_cpu6=time.process_time()
    print("The return array is", A)
    print("the program execution time is", end_exec6-start_exec6)
    print("the cpu execution time is", end_cpu6-start_cpu6)

    print("--(scal)(2, t=(j for j in a))")
    a=np.arange(N)
    start_exec7=time.time()
    start_cpu7=time.process_time()
    starpu.joblib.Parallel(mode="normal", n_jobs=-1, perfmodel="scal")(starpu.joblib.delayed(scal)(2, t=(j for j in a)))
    end_exec7=time.time()
    end_cpu7=time.process_time()
    print("the program execution time is", end_exec7-start_exec7)
    print("the cpu execution time is", end_cpu7-start_cpu7)

    print("--(scal_np)(2,A)")
    A=np.arange(N)
    print("The input is", A)
    start_exec8=time.time()
    start_cpu8=time.process_time()
    starpu.joblib.Parallel(mode="normal", n_jobs=-1, perfmodel="scal")(starpu.joblib.delayed(scal_np)(2,A))
    end_exec8=time.time()
    end_cpu8=time.process_time()
    print("The return array is", A)
    print("the program execution time is", end_exec8-start_exec8)
    print("the cpu execution time is", end_cpu8-start_cpu8)

    print("--(add_scal)(t1=A,t2=B,a=2)")
    A=np.arange(N)
    B=np.arange(N)
    print("The input arrays are t1", A, "t2", B)
    start_exec9=time.time()
    start_cpu9=time.process_time()
    starpu.joblib.Parallel(mode="normal", n_jobs=-1, perfmodel="add_scal")(starpu.joblib.delayed(add_scal)(t1=A,t2=B,a=2))
    end_exec9=time.time()
    end_cpu9=time.process_time()
    print("The return array is", A)
    print("the program execution time is", end_exec9-start_exec9)
    print("the cpu execution time is", end_cpu9-start_cpu9)


    print("--input is iterable function list")
    start_exec10=time.time()
    start_cpu10=time.process_time()
    starpu.joblib.Parallel(mode="normal", n_jobs=-1, perfmodel="func")(g_func)
    end_exec10=time.time()
    end_cpu10=time.process_time()
    print("the program execution time is", end_exec10-start_exec10)
    print("the cpu execution time is", end_cpu10-start_cpu10)

    # def producer():
    # 	for i in range(6):
    # 		print('Produced %s' % i)
    # 		yield i
    #starpu.joblib.Parallel(n_jobs=2)(starpu.joblib.delayed(sqrt)(i) for i in producer())

    print("************************")
    print("parallel Future version:")
    print("************************")
    async def main():

        print("--(sqrt)(i**2)for i in range(N)")
        fut1=starpu.joblib.Parallel(mode="future", n_jobs=-1, perfmodel="sqrt")(starpu.joblib.delayed(sqrt)(i**2)for i in range(N))
        res1=await(fut1)
        print("The result is", sum(res1,[]))

        print("--(multi)(i,j) for i,j in zip(a,b)")
        a=np.arange(N)
        b=np.arange(N, 2*N, 1)
        print("The inputs are a", a, "b", b)
        fut2=starpu.joblib.Parallel(mode="future", n_jobs=-1, perfmodel="multi")(starpu.joblib.delayed(multi)(i,j) for i,j in zip(a,b))
        res2=await(fut2)
        print("The result is", sum(res2,[]))

        print("--(scal_arr)((i for i in b), A)")
        A=np.arange(N)
        b=np.arange(N, 2*N, 1)
        print("The input arrays are A", A, "b", b)
        fut3=starpu.joblib.Parallel(mode="future", n_jobs=-1, perfmodel="scal_arr")(starpu.joblib.delayed(scal_arr)((i for i in b), A))
        res3=await(fut3)
        #print("The return array is", np.concatenate(res3))
        print("The return array is", A)

        print("--(multi_list)((i,j) for i,j in zip(a,b))")
        a=np.arange(N)
        b=np.arange(N, 2*N, 1)
        print("The input lists are a", a, "b", b)
        fut4=starpu.joblib.Parallel(mode="future", n_jobs=-1, perfmodel="multi_list")(starpu.joblib.delayed(multi_list)((i,j) for i,j in zip(a,b)))
        res4=await(fut4)
        print("The result is", sum(res4,[]))

        print("--(multi_2arr)((i for i in a), (j for j in b))")
        a=np.arange(N)
        b=np.arange(N, 2*N, 1)
        print("The input lists are a", a, "b", b)
        fut5=starpu.joblib.Parallel(mode="future", n_jobs=-1, perfmodel="multi_2arr")(starpu.joblib.delayed(multi_2arr)((i for i in a), (j for j in b)))
        res5=await(fut5)
        print("The result is", sum(res5,[]))

        print("--(multi_2np)(b=B, a=A)")
        A=np.arange(N)
        B=np.arange(N, 2*N, 1)
        print("The input arrays are A", A, "B", B)
        fut6=starpu.joblib.Parallel(mode="future", n_jobs=-1, perfmodel="multi_2arr")(starpu.joblib.delayed(multi_2np)(b=B, a=A))
        res6=await(fut6)
        #print("The return array is", np.concatenate(res6))
        print("The return array is", A)


        print("--(scal)(2, (j for j in a))")
        a=np.arange(N)
        print("The input list is a", a)
        fut7=starpu.joblib.Parallel(mode="future", n_jobs=-1, perfmodel="scal")(starpu.joblib.delayed(scal)(2, (j for j in a)))
        res7=await(fut7)
        print("The result is", sum(res7,[]))

        print("--(scal_np)(2,t=A)")
        A=np.arange(N)
        print("The input array is", A)
        fut8=starpu.joblib.Parallel(mode="future", n_jobs=-1, perfmodel="scal")(starpu.joblib.delayed(scal_np)(2,t=A))
        res8=await(fut8)
        #print("The return array is", np.concatenate(res8))
        print("The return array is", A)

        print("--(scal)(2,A,B)")
        A=np.arange(N)
        B=np.arange(N)
        print("The input arrays are A", A, "B", B)
        fut9=starpu.joblib.Parallel(mode="future", n_jobs=-1, perfmodel="add_scal")(starpu.joblib.delayed(add_scal)(2,A,B))
        res9=await(fut9)
        #print("The return array is", np.concatenate(res9))
        print("The return array is", A)

        print("--input is iterable function list")
        fut10=starpu.joblib.Parallel(mode="future", n_jobs=-1, perfmodel="func")(g_func)
        res10=await(fut10)
        #print(res9)

    try:
            asyncio.run(main())
    except starpupy.error as e:
            starpu.shutdown()
            exit(77)

    starpu.perfmodel_plot(perfmodel="sqrt",view=displayPlot)
    starpu.perfmodel_plot(perfmodel="multi",view=displayPlot)
    starpu.perfmodel_plot(perfmodel="scal_arr",view=displayPlot)
    starpu.perfmodel_plot(perfmodel="multi_list",view=displayPlot)
    starpu.perfmodel_plot(perfmodel="multi_2arr",view=displayPlot)
    starpu.perfmodel_plot(perfmodel="scal",view=displayPlot)
    starpu.perfmodel_plot(perfmodel="add_scal",view=displayPlot)
    starpu.perfmodel_plot(perfmodel="func",view=displayPlot)

    starpu.perfmodel_plot(perfmodel="log_list",view=displayPlot)
    starpu.perfmodel_plot(perfmodel="log_arr",view=displayPlot)

starpu.shutdown()
